# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any, AsyncIterator, Optional

from ._ffi_client import FfiClient, FfiHandle
from ._proto import ffi_pb2 as proto_ffi
from ._proto import video_frame_pb2 as proto_video_frame
from ._proto.track_pb2 import TrackSource
from ._utils import RingQueue, task_done_logger
from .participant import Participant
from .track import Track
from .video_frame import VideoFrame


@dataclass
class VideoFrameEvent:
    frame: VideoFrame
    timestamp_us: int
    rotation: proto_video_frame.VideoRotation


class VideoStream:
    """VideoStream is a stream of video frames received from a RemoteTrack."""

    def __init__(
        self,
        track: Track,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        capacity: int = 0,
        format: Optional[proto_video_frame.VideoBufferType.ValueType] = None,
        **kwargs,
    ) -> None:
        self._loop = loop or asyncio.get_event_loop()
        self._ffi_queue = FfiClient.instance.queue.subscribe(self._loop)
        self._queue: RingQueue[VideoFrameEvent | None] = RingQueue(capacity)
        self._track: Track | None = track
        self._format = format
        self._capacity = capacity
        self._format = format
        stream: Any = None
        if "participant" in kwargs:
            stream = self._create_owned_stream_from_participant(
                participant=kwargs["participant"], track_source=kwargs["track_source"]
            )
        else:
            stream = self._create_owned_stream()

        self._ffi_handle = FfiHandle(stream.handle.id)
        self._info = stream.info

        self._task = self._loop.create_task(self._run())
        self._task.add_done_callback(task_done_logger)

    @classmethod
    def from_participant(
        cls,
        *,
        participant: Participant,
        track_source: TrackSource.ValueType,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        format: Optional[proto_video_frame.VideoBufferType.ValueType] = None,
        capacity: int = 0,
    ) -> VideoStream:
        return VideoStream(
            participant=participant,
            track_source=track_source,
            loop=loop,
            capacity=capacity,
            format=format,
            track=None,  # type: ignore
        )

    @classmethod
    def from_track(
        cls,
        *,
        track: Track,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        format: Optional[proto_video_frame.VideoBufferType.ValueType] = None,
        capacity: int = 0,
    ) -> VideoStream:
        return VideoStream(
            track=track,
            loop=loop,
            capacity=capacity,
            format=format,
        )

    def __del__(self) -> None:
        FfiClient.instance.queue.unsubscribe(self._ffi_queue)

    def _create_owned_stream(self) -> Any:
        assert self._track is not None
        req = proto_ffi.FfiRequest()
        new_video_stream = req.new_video_stream
        new_video_stream.track_handle = self._track._ffi_handle.handle
        new_video_stream.type = proto_video_frame.VideoStreamType.VIDEO_STREAM_NATIVE
        if self._format is not None:
            new_video_stream.format = self._format
        new_video_stream.normalize_stride = True
        resp = FfiClient.instance.request(req)
        return resp.new_video_stream.stream

    def _create_owned_stream_from_participant(
        self, participant: Participant, track_source: TrackSource.ValueType
    ) -> Any:
        req = proto_ffi.FfiRequest()
        video_stream_from_participant = req.video_stream_from_participant
        video_stream_from_participant.participant_handle = participant._ffi_handle.handle
        video_stream_from_participant.type = proto_video_frame.VideoStreamType.VIDEO_STREAM_NATIVE
        video_stream_from_participant.track_source = track_source
        video_stream_from_participant.normalize_stride = True
        if self._format is not None:
            video_stream_from_participant.format = self._format
        resp = FfiClient.instance.request(req)
        return resp.video_stream_from_participant.stream

    async def _run(self) -> None:
        while True:
            event = await self._ffi_queue.wait_for(self._is_event)
            video_event = event.video_stream_event

            if video_event.HasField("frame_received"):
                owned_buffer_info = video_event.frame_received.buffer
                frame = VideoFrame._from_owned_info(owned_buffer_info)

                event = VideoFrameEvent(
                    frame=frame,
                    timestamp_us=video_event.frame_received.timestamp_us,
                    rotation=video_event.frame_received.rotation,
                )

                self._queue.put(event)
            elif video_event.HasField("eos"):
                break

        FfiClient.instance.queue.unsubscribe(self._ffi_queue)

    async def aclose(self) -> None:
        self._ffi_handle.dispose()
        await self._task

    def _is_event(self, e: proto_ffi.FfiEvent) -> bool:
        return e.video_stream_event.stream_handle == self._ffi_handle.handle

    def __aiter__(self) -> AsyncIterator[VideoFrameEvent]:
        return self

    async def __anext__(self) -> VideoFrameEvent:
        if self._task.done():
            raise StopAsyncIteration

        item = await self._queue.get()
        if item is None:
            raise StopAsyncIteration

        return item



from __future__ import annotations

from .audio_frame import AudioFrame


__all__ = ["combine_audio_frames"]


def combine_audio_frames(buffer: AudioFrame | list[AudioFrame]) -> AudioFrame:
    """
    Combines one or more `rtc.AudioFrame` objects into a single `rtc.AudioFrame`.

    This function concatenates the audio data from multiple frames, ensuring that
    all frames have the same sample rate and number of channels. It efficiently
    merges the data by preallocating the necessary memory and copying the frame
    data without unnecessary reallocations.

    Args:
        buffer: A single `rtc.AudioFrame` or a list of `rtc.AudioFrame`
            objects to be combined.

    Returns:
        rtc.AudioFrame: A new `rtc.AudioFrame` containing the combined audio data.

    Raises:
        ValueError: If the buffer is empty.
        ValueError: If frames have differing sample rates.
        ValueError: If frames have differing numbers of channels.

    Example:
        >>> frame1 = rtc.AudioFrame(
        ...     data=b"\x01\x02", sample_rate=48000, num_channels=2, samples_per_channel=1
        ... )
        >>> frame2 = rtc.AudioFrame(
        ...     data=b"\x03\x04", sample_rate=48000, num_channels=2, samples_per_channel=1
        ... )
        >>> combined_frame = combine_audio_frames([frame1, frame2])
        >>> combined_frame.data
        b'\x01\x02\x03\x04'
        >>> combined_frame.sample_rate
        48000
        >>> combined_frame.num_channels
        2
        >>> combined_frame.samples_per_channel
        2
    """
    if not isinstance(buffer, list):
        return buffer

    if not buffer:
        raise ValueError("buffer is empty")

    sample_rate = buffer[0].sample_rate
    num_channels = buffer[0].num_channels

    total_data_length = 0
    total_samples_per_channel = 0

    for frame in buffer:
        if frame.sample_rate != sample_rate:
            raise ValueError(
                f"Sample rate mismatch: expected {sample_rate}, got {frame.sample_rate}"
            )

        if frame.num_channels != num_channels:
            raise ValueError(
                f"Channel count mismatch: expected {num_channels}, got {frame.num_channels}"
            )

        total_data_length += len(frame.data)
        total_samples_per_channel += frame.samples_per_channel

    data = bytearray(total_data_length)
    offset = 0
    for frame in buffer:
        frame_data = frame.data.cast("b")
        data[offset : offset + len(frame_data)] = frame_data
        offset += len(frame_data)

    return AudioFrame(
        data=data,
        sample_rate=sample_rate,
        num_channels=num_channels,
        samples_per_channel=total_samples_per_channel,
    )



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import ctypes
import asyncio
import os
import mimetypes
import aiofiles
from typing import List, Union, Callable, Dict, Awaitable, Optional, Mapping, cast, TypeVar
from abc import abstractmethod, ABC

from ._ffi_client import FfiClient, FfiHandle
from ._proto import ffi_pb2 as proto_ffi
from ._proto import participant_pb2 as proto_participant
from ._proto.room_pb2 import (
    TrackPublishOptions,
)
from ._proto.room_pb2 import (
    TranscriptionSegment as ProtoTranscriptionSegment,
)
from ._proto.track_pb2 import (
    ParticipantTrackPermission,
)
from ._utils import BroadcastQueue
from .track import LocalTrack
from .track_publication import (
    LocalTrackPublication,
    RemoteTrackPublication,
    TrackPublication,
)
from .transcription import Transcription
from .rpc import RpcError
from ._proto.rpc_pb2 import RpcMethodInvocationResponseRequest
from .log import logger

from .rpc import RpcInvocationData
from .data_stream import (
    TextStreamWriter,
    ByteStreamWriter,
    ByteStreamInfo,
    STREAM_CHUNK_SIZE,
)


class PublishTrackError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class UnpublishTrackError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class PublishDataError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class PublishDTMFError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class PublishTranscriptionError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class Participant(ABC):
    def __init__(self, owned_info: proto_participant.OwnedParticipant) -> None:
        self._info = owned_info.info
        self._ffi_handle = FfiHandle(owned_info.handle.id)

    @property
    @abstractmethod
    def track_publications(self) -> Mapping[str, TrackPublication]:
        """
        A dictionary of track publications associated with the participant.
        """
        ...

    @property
    def sid(self) -> str:
        return self._info.sid

    @property
    def name(self) -> str:
        return self._info.name

    @property
    def identity(self) -> str:
        return self._info.identity

    @property
    def metadata(self) -> str:
        return self._info.metadata

    @property
    def attributes(self) -> dict[str, str]:
        """Custom attributes associated with the participant."""
        return dict(self._info.attributes)

    @property
    def kind(self) -> proto_participant.ParticipantKind.ValueType:
        """Participant's kind (e.g., regular participant, ingress, egress, sip, agent)."""
        return self._info.kind

    @property
    def disconnect_reason(
        self,
    ) -> Optional[proto_participant.DisconnectReason.ValueType]:
        """Reason for the participant's disconnection.

        Returns one of DisconnectReasons or None if the participant isn't disconnected. Common reasons are:
        - CLIENT_INITIATED - the client initiated the disconnect
        - DUPLICATE_IDENTITY - another participant with the same identity has joined the room
        - SERVER_SHUTDOWN - the server instance is shutting down
        - PARTICIPANT_REMOVED - RoomService.RemoveParticipant was called
        - ROOM_DELETED - RoomService.DeleteRoom was called
        - STATE_MISMATCH - the client is attempting to resume a session, but server is not aware of it
        - JOIN_FAILURE - client was unable to connect fully

        When dialing a participant via SIP, you may see the following reasons:
        - USER_UNAVAILABLE - SIP callee did not respond in time
        - USER_REJECTED - SIP callee rejected the call (busy)
        - SIP_TRUNK_FAILURE - SIP protocol failure or unexpected response
        """
        if self._info.disconnect_reason == proto_participant.DisconnectReason.UNKNOWN_REASON:
            return None
        return self._info.disconnect_reason


RpcHandler = Callable[["RpcInvocationData"], Union[Awaitable[Optional[str]], Optional[str]]]
F = TypeVar(
    "F", bound=Callable[[RpcInvocationData], Union[Awaitable[Optional[str]], Optional[str]]]
)


class LocalParticipant(Participant):
    """Represents the local participant in a room."""

    def __init__(
        self,
        room_queue: BroadcastQueue[proto_ffi.FfiEvent],
        owned_info: proto_participant.OwnedParticipant,
    ) -> None:
        super().__init__(owned_info)
        self._room_queue = room_queue
        self._track_publications: dict[str, LocalTrackPublication] = {}  # type: ignore
        self._rpc_handlers: Dict[str, RpcHandler] = {}

    @property
    def track_publications(self) -> Mapping[str, LocalTrackPublication]:
        """
        A dictionary of track publications associated with the participant.
        """
        return self._track_publications

    async def publish_data(
        self,
        payload: Union[bytes, str],
        *,
        reliable: bool = True,
        destination_identities: List[str] = [],
        topic: str = "",
    ) -> None:
        """
        Publish arbitrary data to the room.

        Args:
            payload (Union[bytes, str]): The data to publish.
            reliable (bool, optional): Whether to send reliably or not. Defaults to True.
            destination_identities (List[str], optional): List of participant identities to send to. Defaults to [].
            topic (str, optional): The topic under which to publish the data. Defaults to "".

        Raises:
            PublishDataError: If there is an error in publishing data.
        """
        if isinstance(payload, str):
            payload = payload.encode("utf-8")

        data_len = len(payload)
        cdata = (ctypes.c_byte * data_len)(*payload)

        req = proto_ffi.FfiRequest()
        req.publish_data.local_participant_handle = self._ffi_handle.handle
        req.publish_data.data_ptr = ctypes.addressof(cdata)
        req.publish_data.data_len = data_len
        req.publish_data.reliable = reliable
        req.publish_data.topic = topic
        req.publish_data.destination_identities.extend(destination_identities)

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.publish_data.async_id == resp.publish_data.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.publish_data.error:
            raise PublishDataError(cb.publish_data.error)

    async def publish_dtmf(self, *, code: int, digit: str) -> None:
        """
        Publish SIP DTMF message.

        Args:
            code (int): DTMF code.
            digit (str): DTMF digit.

        Raises:
            PublishDTMFError: If there is an error in publishing SIP DTMF message.
        """
        req = proto_ffi.FfiRequest()
        req.publish_sip_dtmf.local_participant_handle = self._ffi_handle.handle
        req.publish_sip_dtmf.code = code
        req.publish_sip_dtmf.digit = digit

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.publish_sip_dtmf.async_id == resp.publish_sip_dtmf.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.publish_sip_dtmf.error:
            raise PublishDTMFError(cb.publish_sip_dtmf.error)

    async def publish_transcription(self, transcription: Transcription) -> None:
        """
        Publish transcription data to the room.

        Args:
            transcription (Transcription): The transcription data to publish.

        Raises:
            PublishTranscriptionError: If there is an error in publishing transcription.
        """
        req = proto_ffi.FfiRequest()
        proto_segments = [
            ProtoTranscriptionSegment(
                id=s.id,
                text=s.text,
                start_time=s.start_time,
                end_time=s.end_time,
                final=s.final,
                language=s.language,
            )
            for s in transcription.segments
        ]
        # fmt: off
        req.publish_transcription.local_participant_handle = self._ffi_handle.handle
        req.publish_transcription.participant_identity = transcription.participant_identity
        req.publish_transcription.segments.extend(proto_segments)
        req.publish_transcription.track_id = transcription.track_sid
        # fmt: on
        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.publish_transcription.async_id == resp.publish_transcription.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.publish_transcription.error:
            raise PublishTranscriptionError(cb.publish_transcription.error)

    async def perform_rpc(
        self,
        *,
        destination_identity: str,
        method: str,
        payload: str,
        response_timeout: Optional[float] = None,
    ) -> str:
        """
        Initiate an RPC call to a remote participant.

        Args:
            destination_identity (str): The `identity` of the destination participant
            method (str): The method name to call
            payload (str): The method payload
            response_timeout (Optional[float]): Timeout for receiving a response after initial connection

        Returns:
            str: The response payload

        Raises:
            RpcError: On failure. Details in `message`.
        """
        req = proto_ffi.FfiRequest()
        req.perform_rpc.local_participant_handle = self._ffi_handle.handle
        req.perform_rpc.destination_identity = destination_identity
        req.perform_rpc.method = method
        req.perform_rpc.payload = payload
        if response_timeout is not None:
            req.perform_rpc.response_timeout_ms = int(response_timeout * 1000)

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb = await queue.wait_for(
                lambda e: (e.perform_rpc.async_id == resp.perform_rpc.async_id)
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

        if cb.perform_rpc.HasField("error"):
            raise RpcError._from_proto(cb.perform_rpc.error)

        return cb.perform_rpc.payload

    def register_rpc_method(
        self,
        method_name: str,
        handler: Optional[F] = None,
    ) -> Union[F, Callable[[F], F]]:
        """
        Establishes the participant as a receiver for calls of the specified RPC method.
        Can be used either as a decorator or a regular method.

        The handler will receive one argument of type `RpcInvocationData` and should return a string response which will be forwarded back to the caller.

        The handler may be synchronous or asynchronous.

        If unable to respond within `response_timeout`, the caller will hang up and receive an error on their side.

        You may raise errors of type `RpcError` in the handler, and they will be forwarded to the caller.

        Other errors raised in your handler will be caught and forwarded to the caller as "1500 Application Error".

        Args:
            method_name (str): The name of the indicated RPC method.
            handler (Optional[Callable]): Handler to be invoked whenever an RPC request for this method is received.  Omit this argument to use the decorator syntax.

        Returns:
            None (when used as a decorator it returns the decorator function)

        Example:
            # As a decorator:
            @room.local_participant.register_rpc_method("greet")
            async def greet_handler(data: RpcInvocationData) -> str:
                print(f"Received greeting from {data.caller_identity}: {data.payload}")
                return f"Hello, {data.caller_identity}!"

            # As a regular method:
            async def greet_handler(data: RpcInvocationData) -> str:
                print(f"Received greeting from {data.caller_identity}: {data.payload}")
                return f"Hello, {data.caller_identity}!"

            room.local_participant.register_rpc_method('greet', greet_handler)
        """

        def register(handler_func: F) -> F:
            self._rpc_handlers[method_name] = handler_func
            req = proto_ffi.FfiRequest()
            req.register_rpc_method.local_participant_handle = self._ffi_handle.handle
            req.register_rpc_method.method = method_name
            FfiClient.instance.request(req)
            return handler_func

        if handler is not None:
            return register(handler)
        else:
            return register

    def unregister_rpc_method(self, method: str) -> None:
        """
        Unregisters a previously registered RPC method.

        Args:
            method (str): The name of the RPC method to unregister
        """
        self._rpc_handlers.pop(method, None)

        req = proto_ffi.FfiRequest()
        req.unregister_rpc_method.local_participant_handle = self._ffi_handle.handle
        req.unregister_rpc_method.method = method

        FfiClient.instance.request(req)

    def set_track_subscription_permissions(
        self,
        *,
        allow_all_participants: bool,
        participant_permissions: Optional[List[ParticipantTrackPermission]] = None,
    ) -> None:
        """
        Set the track subscription permissions for the local participant.

        Args:
            allow_all_participants (bool): Whether to allow all participants to subscribe to this participant's tracks.
            participant_permissions (List[ParticipantTrackPermission]): Participant-specific track subscription permissions, ignored if `allow_all_participants` is True.
        """
        if participant_permissions is None:
            participant_permissions = []

        req = proto_ffi.FfiRequest()
        req.set_track_subscription_permissions.local_participant_handle = self._ffi_handle.handle
        req.set_track_subscription_permissions.all_participants_allowed = allow_all_participants
        req.set_track_subscription_permissions.permissions.extend(participant_permissions)
        FfiClient.instance.request(req)

    async def _handle_rpc_method_invocation(
        self,
        invocation_id: int,
        method: str,
        request_id: str,
        caller_identity: str,
        payload: str,
        response_timeout: float,
    ) -> None:
        response_error: Optional[RpcError] = None
        response_payload: Optional[str] = None

        params = RpcInvocationData(request_id, caller_identity, payload, response_timeout)

        handler = self._rpc_handlers.get(method)

        if not handler:
            response_error = RpcError._built_in(RpcError.ErrorCode.UNSUPPORTED_METHOD)
        else:
            try:
                if asyncio.iscoroutinefunction(handler):
                    try:
                        response_payload = await asyncio.wait_for(
                            handler(params), timeout=response_timeout
                        )
                    except asyncio.TimeoutError:
                        raise RpcError._built_in(RpcError.ErrorCode.RESPONSE_TIMEOUT)
                    except asyncio.CancelledError:
                        raise RpcError._built_in(RpcError.ErrorCode.RECIPIENT_DISCONNECTED)
                else:
                    response_payload = cast(Optional[str], handler(params))
            except RpcError as error:
                response_error = error
            except Exception:
                logger.exception(
                    f"Uncaught error returned by RPC handler for {method}. "
                    "Returning APPLICATION_ERROR instead. "
                )
                response_error = RpcError._built_in(RpcError.ErrorCode.APPLICATION_ERROR)

        req = proto_ffi.FfiRequest(
            rpc_method_invocation_response=RpcMethodInvocationResponseRequest(
                local_participant_handle=self._ffi_handle.handle,
                invocation_id=invocation_id,
                error=response_error._to_proto() if response_error else None,
                payload=response_payload,
            )
        )

        res = FfiClient.instance.request(req)

        if res.rpc_method_invocation_response.error:
            err = res.rpc_method_invocation_response.error
            logger.error(f"error sending rpc method invocation response: {err}")

    async def set_metadata(self, metadata: str) -> None:
        """
        Set the metadata for the local participant.

        Note: this requires `canUpdateOwnMetadata` permission.

        Args:
            metadata (str): The new metadata.
        """
        req = proto_ffi.FfiRequest()
        req.set_local_metadata.local_participant_handle = self._ffi_handle.handle
        req.set_local_metadata.metadata = metadata

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            await queue.wait_for(
                lambda e: e.set_local_metadata.async_id == resp.set_local_metadata.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

    async def set_name(self, name: str) -> None:
        """
        Set the name for the local participant.

        Note: this requires `canUpdateOwnMetadata` permission.

        Args:
            name (str): The new name.
        """
        req = proto_ffi.FfiRequest()
        req.set_local_name.local_participant_handle = self._ffi_handle.handle
        req.set_local_name.name = name

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            await queue.wait_for(
                lambda e: e.set_local_name.async_id == resp.set_local_name.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

    async def set_attributes(self, attributes: dict[str, str]) -> None:
        """
        Set custom attributes for the local participant.

        Note: this requires `canUpdateOwnMetadata` permission.

        Args:
            attributes (dict[str, str]): A dictionary of attributes to set.
        """
        req = proto_ffi.FfiRequest()
        req.set_local_attributes.local_participant_handle = self._ffi_handle.handle
        existing_attributes = {
            entry.key: entry.value for entry in req.set_local_attributes.attributes
        }
        existing_attributes.update(attributes)

        for key, value in existing_attributes.items():
            entry = req.set_local_attributes.attributes.add()
            entry.key = key
            entry.value = value

        queue = FfiClient.instance.queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            await queue.wait_for(
                lambda e: e.set_local_attributes.async_id == resp.set_local_attributes.async_id
            )
        finally:
            FfiClient.instance.queue.unsubscribe(queue)

    async def stream_text(
        self,
        *,
        destination_identities: Optional[List[str]] = None,
        topic: str = "",
        attributes: Optional[Dict[str, str]] = None,
        stream_id: str | None = None,
        reply_to_id: str | None = None,
        total_size: int | None = None,
        sender_identity: str | None = None,
    ) -> TextStreamWriter:
        """
        Returns a TextStreamWriter that allows to write individual chunks of text to a text stream.
        In most cases where you want to simply send a text message use send_text() instead.
        """
        writer = TextStreamWriter(
            self,
            topic=topic,
            attributes=attributes,
            reply_to_id=reply_to_id,
            destination_identities=destination_identities,
            total_size=total_size,
            stream_id=stream_id,
            sender_identity=sender_identity,
        )

        await writer._send_header()

        return writer

    async def send_text(
        self,
        text: str,
        *,
        destination_identities: Optional[List[str]] = None,
        topic: str = "",
        attributes: Optional[Dict[str, str]] = None,
        reply_to_id: str | None = None,
    ):
        total_size = len(text.encode())
        writer = await self.stream_text(
            destination_identities=destination_identities,
            topic=topic,
            attributes=attributes,
            reply_to_id=reply_to_id,
            total_size=total_size,
        )

        await writer.write(text)
        await writer.aclose()

        return writer.info

    async def stream_bytes(
        self,
        name: str,
        *,
        total_size: int | None = None,
        mime_type: str = "application/octet-stream",
        attributes: Optional[Dict[str, str]] = None,
        stream_id: str | None = None,
        destination_identities: Optional[List[str]] = None,
        topic: str = "",
    ) -> ByteStreamWriter:
        """
        Returns a ByteStreamWriter that allows to write individual chunks of bytes to a byte stream.
        In cases where you want to simply send a file from the file system use send_file() instead.
        """
        writer = ByteStreamWriter(
            self,
            name=name,
            attributes=attributes,
            total_size=total_size,
            stream_id=stream_id,
            mime_type=mime_type,
            destination_identities=destination_identities,
            topic=topic,
        )

        await writer._send_header()

        return writer

    async def send_file(
        self,
        file_path: str,
        *,
        topic: str = "",
        destination_identities: Optional[List[str]] = None,
        attributes: Optional[Dict[str, str]] = None,
        stream_id: str | None = None,
    ) -> ByteStreamInfo:
        file_size = os.path.getsize(file_path)
        file_name = os.path.basename(file_path)
        mime_type, _ = mimetypes.guess_type(file_path)
        if mime_type is None:
            mime_type = "application/octet-stream"  # Fallback MIME type for unknown files

        writer: ByteStreamWriter = await self.stream_bytes(
            name=file_name,
            total_size=file_size,
            mime_type=mime_type,
            attributes=attributes,
            stream_id=stream_id,
            destination_identities=destination_identities,
            topic=topic,
        )

        async with aiofiles.open(file_path, "rb") as f:
            while bytes := await f.read(STREAM_CHUNK_SIZE):
                await writer.write(bytes)
        await writer.aclose()

        return writer.info

    async def publish_track(
        self, track: LocalTrack, options: TrackPublishOptions = TrackPublishOptions()
    ) -> LocalTrackPublication:
        """
        Publish a local track to the room.

        Args:
            track (LocalTrack): The track to publish.
            options (TrackPublishOptions, optional): Options for publishing the track.

        Returns:
            LocalTrackPublication: The publication of the published track.

        Raises:
            PublishTrackError: If there is an error in publishing the track.
        """
        req = proto_ffi.FfiRequest()
        req.publish_track.track_handle = track._ffi_handle.handle
        req.publish_track.local_participant_handle = self._ffi_handle.handle
        req.publish_track.options.CopyFrom(options)

        queue = self._room_queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.publish_track.async_id == resp.publish_track.async_id
            )

            if cb.publish_track.error:
                raise PublishTrackError(cb.publish_track.error)

            track_publication = LocalTrackPublication(cb.publish_track.publication)
            track_publication._track = track
            track._info.sid = track_publication.sid
            self._track_publications[track_publication.sid] = track_publication

            queue.task_done()
            return track_publication
        finally:
            self._room_queue.unsubscribe(queue)

    async def unpublish_track(self, track_sid: str) -> None:
        """
        Unpublish a track from the room.

        Args:
            track_sid (str): The SID of the track to unpublish.

        Raises:
            UnpublishTrackError: If there is an error in unpublishing the track.
        """
        req = proto_ffi.FfiRequest()
        req.unpublish_track.local_participant_handle = self._ffi_handle.handle
        req.unpublish_track.track_sid = track_sid
        req.unpublish_track.stop_on_unpublish = True

        queue = self._room_queue.subscribe()
        try:
            resp = FfiClient.instance.request(req)
            cb: proto_ffi.FfiEvent = await queue.wait_for(
                lambda e: e.unpublish_track.async_id == resp.unpublish_track.async_id
            )

            if cb.unpublish_track.error:
                raise UnpublishTrackError(cb.unpublish_track.error)

            publication = self._track_publications.pop(track_sid)
            publication._track = None
            queue.task_done()
        finally:
            self._room_queue.unsubscribe(queue)

    def __repr__(self) -> str:
        return f"rtc.LocalParticipant(sid={self.sid}, identity={self.identity}, name={self.name})"


class RemoteParticipant(Participant):
    def __init__(self, owned_info: proto_participant.OwnedParticipant) -> None:
        super().__init__(owned_info)
        self._track_publications: dict[str, RemoteTrackPublication] = {}  # type: ignore

    @property
    def track_publications(self) -> Mapping[str, RemoteTrackPublication]:
        """
        A dictionary of track publications associated with the participant.
        """
        return self._track_publications

    def __repr__(self) -> str:
        return f"rtc.RemoteParticipant(sid={self.sid}, identity={self.identity}, name={self.name})"



import inspect
import asyncio
from typing import Callable, Dict, Set, Optional, Generic, TypeVar

from .log import logger

T_contra = TypeVar("T_contra", contravariant=True)


class EventEmitter(Generic[T_contra]):
    def __init__(self) -> None:
        """
        Initialize a new instance of EventEmitter.
        """
        self._events: Dict[T_contra, Set[Callable]] = dict()

    def emit(self, event: T_contra, *args) -> None:
        """
        Trigger all callbacks associated with the given event.

        Args:
            event (T): The event to emit.
            *args: Positional arguments to pass to the callbacks.

        Example:
            Basic usage of emit:

            ```python
            emitter = EventEmitter[str]()

            def greet(name):
                print(f"Hello, {name}!")

            emitter.on('greet', greet)
            emitter.emit('greet', 'Alice')  # Output: Hello, Alice!
            ```
        """
        if event in self._events:
            callables = self._events[event].copy()
            for callback in callables:
                try:
                    sig = inspect.signature(callback)
                    params = sig.parameters.values()

                    has_varargs = any(p.kind == p.VAR_POSITIONAL for p in params)
                    if has_varargs:
                        callback(*args)
                    else:
                        positional_params = [
                            p
                            for p in params
                            if p.kind in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD)
                        ]
                        num_params = len(positional_params)
                        num_args = min(len(args), num_params)
                        callback_args = args[:num_args]

                        callback(*callback_args)
                except TypeError:
                    raise
                except Exception:
                    logger.exception(f"failed to emit event {event}")

    def once(self, event: T_contra, callback: Optional[Callable] = None) -> Callable:
        """
        Register a callback to be called only once when the event is emitted.

        If a callback is provided, it registers the callback directly.
        If no callback is provided, it returns a decorator for use with function definitions.

        Args:
            event (T): The event to listen for.
            callback (Callable, optional): The callback to register. Defaults to None.

        Returns:
            Callable: The registered callback or a decorator if callback is None.

        Example:
            Using once with a direct callback:

            ```python
            emitter = EventEmitter[str]()

            def greet_once(name):
                print(f"Hello once, {name}!")

            emitter.once('greet', greet_once)
            emitter.emit('greet', 'Bob')    # Output: Hello once, Bob!
            emitter.emit('greet', 'Bob')    # No output, callback was removed after first call
            ```

            Using once as a decorator:

            ```python
            emitter = EventEmitter[str]()

            @emitter.once('greet')
            def greet_once(name):
                print(f"Hello once, {name}!")

            emitter.emit('greet', 'Bob')    # Output: Hello once, Bob!
            emitter.emit('greet', 'Bob')    # No output
            ```
        """
        if callback is not None:

            def once_callback(*args, **kwargs):
                self.off(event, once_callback)
                callback(*args, **kwargs)

            return self.on(event, once_callback)
        else:

            def decorator(callback: Callable) -> Callable:
                self.once(event, callback)
                return callback

            return decorator

    def on(self, event: T_contra, callback: Optional[Callable] = None) -> Callable:
        """
        Register a callback to be called whenever the event is emitted.

        If a callback is provided, it registers the callback directly.
        If no callback is provided, it returns a decorator for use with function definitions.

        Args:
            event (T): The event to listen for.
            callback (Callable, optional): The callback to register. Defaults to None.

        Returns:
            Callable: The registered callback or a decorator if callback is None.

        Example:
            Using on with a direct callback:

            ```python
            emitter = EventEmitter[str]()

            def greet(name):
                print(f"Hello, {name}!")

            emitter.on('greet', greet)
            emitter.emit('greet', 'Charlie')  # Output: Hello, Charlie!
            ```

            Using on as a decorator:

            ```python
            emitter = EventEmitter[str]()

            @emitter.on('greet')
            def greet(name):
                print(f"Hello, {name}!")

            emitter.emit('greet', 'Charlie')  # Output: Hello, Charlie!
            ```
        """
        if callback is not None:
            if asyncio.iscoroutinefunction(callback):
                raise ValueError(
                    "Cannot register an async callback with `.on()`. Use `asyncio.create_task` within your synchronous callback instead."
                )

            if event not in self._events:
                self._events[event] = set()
            self._events[event].add(callback)
            return callback
        else:

            def decorator(callback: Callable) -> Callable:
                self.on(event, callback)
                return callback

            return decorator

    def off(self, event: T_contra, callback: Callable) -> None:
        """
        Unregister a callback from an event.

        Args:
            event (T): The event to stop listening to.
            callback (Callable): The callback to remove.

        Example:
            Removing a callback:

            ```python
            emitter = EventEmitter[str]()

            def greet(name):
                print(f"Hello, {name}!")

            emitter.on('greet', greet)
            emitter.off('greet', greet)
            emitter.emit('greet', 'Dave')  # No output, callback was removed
            ```
        """
        if event in self._events:
            self._events[event].discard(callback)



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for recording remote participant streams to WebM files."""

from __future__ import annotations

import asyncio
import logging
import gc
from dataclasses import dataclass
from typing import Optional, Set
from pathlib import Path
from fractions import Fraction
import time

try:
    import av
    HAS_AV = True
except ImportError:
    HAS_AV = False
    av = None  # type: ignore

from .room import Room
from .participant import RemoteParticipant
from .track import RemoteVideoTrack, RemoteAudioTrack, Track
from .track_publication import RemoteTrackPublication
from .video_stream import VideoStream, VideoFrameEvent
from .audio_stream import AudioStream, AudioFrameEvent
from .video_frame import VideoFrame
from .audio_frame import AudioFrame
from ._proto.track_pb2 import TrackKind
from ._proto.video_frame_pb2 import VideoBufferType

logger = logging.getLogger(__name__)


class RecordingError(Exception):
    """Base exception for recording-related errors."""

    pass


class ParticipantNotFoundError(RecordingError):
    """Raised when a participant is not found in the room."""

    pass


class TrackNotFoundError(RecordingError):
    """Raised when a track is not found or not available."""

    pass


class WebMEncoderNotAvailableError(RecordingError):
    """Raised when WebM encoding libraries are not available."""

    pass


@dataclass
class RecordingStats:
    """Statistics for a recording session."""

    video_frames_recorded: int = 0
    audio_frames_recorded: int = 0
    recording_duration_seconds: float = 0.0
    output_file_size_bytes: int = 0


class ParticipantRecorder:
    """Records video and audio from a remote participant to a WebM file.

    This class captures WebRTC streams from a specific remote participant in a LiveKit room
    and encodes them into a WebM file with VP8/VP9 video and Opus audio codecs.

    Example:
        ```python
        recorder = ParticipantRecorder(room)
        await recorder.start_recording("participant_identity")
        
        # Recording is now in progress...
        await asyncio.sleep(60)  # Record for 1 minute
        
        await recorder.stop_recording("output.webm")
        stats = recorder.get_stats()
        print(f"Recorded {stats.video_frames_recorded} video frames")
        ```
    """

    def __init__(
        self,
        room: Room,
        video_codec: str = "vp8",
        audio_codec: str = "opus",
        video_bitrate: int = 2000000,  # 2 Mbps
        audio_bitrate: int = 128000,  # 128 kbps
        video_fps: int = 30,
        video_quality: str = "medium",  # "low", "medium", "high", "best"
        auto_bitrate: bool = True,  # Auto-adjust bitrate based on resolution
    ) -> None:
        """Initialize a ParticipantRecorder instance.

        Args:
            room: The LiveKit Room instance connected to the session.
            video_codec: Video codec to use ('vp8' or 'vp9'). Defaults to 'vp8'.
                VP9 provides better quality at the same bitrate but is slower to encode.
            audio_codec: Audio codec to use. Defaults to 'opus'.
            video_bitrate: Target video bitrate in bits per second. Defaults to 2000000 (2 Mbps).
                If auto_bitrate is True, this will be adjusted based on resolution.
            audio_bitrate: Target audio bitrate in bits per second. Defaults to 128000 (128 kbps).
            video_fps: Target video frame rate. Defaults to 30.
            video_quality: Quality preset for encoding ('low', 'medium', 'high', 'best').
                Higher quality uses slower encoding but produces better results.
                Defaults to 'medium'.
            auto_bitrate: If True, automatically adjust bitrate based on resolution.
                Higher resolutions get higher bitrates. Defaults to True.

        Raises:
            WebMEncoderNotAvailableError: If PyAV is not installed.
        """
        if not HAS_AV:
            raise WebMEncoderNotAvailableError(
                "PyAV is required for recording. Install it with: pip install av"
            )

        if video_codec not in ("vp8", "vp9"):
            raise ValueError("video_codec must be 'vp8' or 'vp9'")
        
        if video_quality not in ("low", "medium", "high", "best"):
            raise ValueError("video_quality must be 'low', 'medium', 'high', or 'best'")

        self.room = room
        self.video_codec = video_codec
        self.audio_codec = audio_codec
        self.video_bitrate = video_bitrate
        self.audio_bitrate = audio_bitrate
        self.video_fps = video_fps
        self.video_quality = video_quality
        self.auto_bitrate = auto_bitrate

        self._participant_identity: Optional[str] = None
        self._participant: Optional[RemoteParticipant] = None
        self._is_recording: bool = False
        self._recording_task: Optional[asyncio.Task[None]] = None

        # Streams for video and audio capture
        self._video_capture_stream: Optional[VideoStream] = None
        self._audio_capture_stream: Optional[AudioStream] = None

        # Frame queues with bounded size to prevent excessive memory growth
        # Buffer approximately 30 seconds of frames at target fps to handle bursts
        # This allows sufficient buffering while still limiting memory growth
        # For 30fps video: 30 * 30 = 900 frames (~27MB of video data)
        # For audio at ~100fps: 30 * 100 = 3000 frames (~10MB of audio data)
        max_video_queue_size = max(500, video_fps * 30)  # At least 500, or 30 seconds worth
        # Audio frames arrive at ~100fps (every ~10ms), so 30 seconds = 3000 frames
        max_audio_queue_size = max(1000, video_fps * 100)  # ~30 seconds at typical audio frame rate
        
        self._video_queue: asyncio.Queue[VideoFrameEvent] = asyncio.Queue(maxsize=max_video_queue_size)
        self._audio_queue: asyncio.Queue[AudioFrameEvent] = asyncio.Queue(maxsize=max_audio_queue_size)

        # Background tasks for capturing frames
        self._video_capture_task: Optional[asyncio.Task[None]] = None
        self._audio_capture_task: Optional[asyncio.Task[None]] = None

        # Incremental encoding state
        self._encoding_task: Optional[asyncio.Task[None]] = None
        self._output_container: Optional[av.container.OutputContainer] = None
        self._output_file_path: Optional[str] = None
        self._container_was_initialized: bool = False  # Track if container was ever initialized
        self._video_stream: Optional[av.VideoStream] = None  # PyAV stream
        self._audio_stream: Optional[av.AudioStream] = None  # PyAV stream
        self._video_stream_initialized: bool = False
        self._audio_stream_initialized: bool = False
        self._cumulative_audio_samples: int = 0
        self._first_video_frame_time: Optional[float] = None
        self._video_frame_count: int = 0

        # Monotonicity enforcement
        self._last_video_dts: int = -1
        self._last_audio_dts: int = -1

        # Synchronization
        self._lock = asyncio.Lock()
        self._start_time: Optional[float] = None
        self._recording_start_time_us: Optional[int] = None

        # Statistics
        self._stats = RecordingStats()

        # Track subscriptions tracking
        self._subscribed_track_sids: Set[str] = set()

    async def start_recording(self, participant_identity: str) -> None:
        """Start recording from the specified participant.

        Args:
            participant_identity: The identity of the remote participant to record.

        Raises:
            ParticipantNotFoundError: If the participant is not found in the room.
            RecordingError: If recording is already in progress or fails to start.
        """
        async with self._lock:
            if self._is_recording:
                raise RecordingError("Recording is already in progress")

            # Find the participant
            participant = self.room.remote_participants.get(participant_identity)
            if not participant:
                raise ParticipantNotFoundError(
                    f"Participant '{participant_identity}' not found in room"
                )

            self._participant_identity = participant_identity
            self._participant = participant
            self._is_recording = True
            self._start_time = time.time()

            # Reset incremental encoding state for new recording
            self._first_video_frame_time = None
            self._cumulative_audio_samples = 0
            self._video_frame_count = 0
            self._last_video_dts = -1
            self._last_audio_dts = -1

            # Subscribe to all published tracks
            await self._subscribe_to_participant_tracks(participant)

            # Set up event handlers for tracks that may be published later
            self._setup_track_handlers(participant)

            # Wait for tracks to be available and start capturing
            await self._wait_for_tracks_and_start_capture()
            
            # Start incremental encoding task
            self._encoding_task = asyncio.create_task(self._incremental_encoding_loop())

            logger.info(
                f"Started recording participant '{participant_identity}'"
            )

    async def _subscribe_to_participant_tracks(
        self, participant: RemoteParticipant
    ) -> None:
        """Subscribe to all published tracks of a participant."""
        for publication in participant.track_publications.values():
            if not publication.subscribed:
                publication.set_subscribed(True)
                self._subscribed_track_sids.add(publication.sid)
                logger.debug(
                    f"Subscribed to track {publication.sid} "
                    f"(kind: {publication.kind})"
                )

    async def _unsubscribe_from_participant_tracks(
        self, participant: Optional[RemoteParticipant]
    ) -> None:
        """Unsubscribe from all tracks that were subscribed by this recorder.
        
        This method only unsubscribes from tracks that this recorder subscribed to,
        to avoid interfering with other subscriptions that may exist.
        """
        if not participant:
            return
        
        for publication in participant.track_publications.values():
            # Only unsubscribe if we subscribed to this track
            if publication.sid in self._subscribed_track_sids and publication.subscribed:
                try:
                    publication.set_subscribed(False)
                    logger.debug(
                        f"Unsubscribed from track {publication.sid} "
                        f"(kind: {publication.kind})"
                    )
                except Exception as e:
                    logger.warning(
                        f"Error unsubscribing from track {publication.sid}: {e}"
                    )

    def _setup_track_handlers(self, participant: RemoteParticipant) -> None:
        """Set up event handlers for track publication events."""
        original_on_track_subscribed = None

        async def on_track_subscribed_wrapper(
            track: Track,
            publication: RemoteTrackPublication,
            p: RemoteParticipant,
        ) -> None:
            # Only handle tracks from the participant we're recording
            if p.identity != self._participant_identity:
                return

            if original_on_track_subscribed:
                await original_on_track_subscribed(track, publication, p)

            # If we're recording, start capturing from this track
            if self._is_recording and publication.sid not in self._subscribed_track_sids:
                self._subscribed_track_sids.add(publication.sid)
                await self._start_capture_for_track(track, publication)

        # This is a simplified approach - in practice, we rely on the room's
        # track_subscribed event which we'll monitor in the main recording loop
        pass

    async def _wait_for_tracks_and_start_capture(self) -> None:
        """Wait for tracks to be subscribed and start capturing frames."""
        if not self._participant:
            return

        # Wait a bit for tracks to become available
        await asyncio.sleep(0.5)

        # Start capturing from already-subscribed tracks
        for publication in self._participant.track_publications.values():
            if publication.subscribed and publication.track:
                track = publication.track
                if publication.sid not in self._subscribed_track_sids:
                    self._subscribed_track_sids.add(publication.sid)
                    await self._start_capture_for_track(track, publication)

        # If no tracks found yet, we'll wait for them via event handlers
        if not self._subscribed_track_sids:
            logger.warning(
                f"No tracks found for participant '{self._participant_identity}'. "
                "Will wait for tracks to be published."
            )

    async def _start_capture_for_track(
        self, track: Track, publication: RemoteTrackPublication
    ) -> None:
        """Start capturing frames from a track."""
        if track.kind == TrackKind.KIND_VIDEO and isinstance(
            track, RemoteVideoTrack
        ):
            if self._video_capture_stream is None:
                # Set bounded capacity on VideoStream's internal queue to prevent unbounded memory growth
                # Use same capacity as our recorder queue (~30 seconds of frames)
                stream_capacity = max(500, self.video_fps * 30)
                self._video_capture_stream = VideoStream(track, capacity=stream_capacity)
                self._video_capture_task = asyncio.create_task(
                    self._capture_video_frames()
                )
                logger.debug(f"Started video capture from track {track.sid} with capacity={stream_capacity}")

        elif track.kind == TrackKind.KIND_AUDIO and isinstance(
            track, RemoteAudioTrack
        ):
            if self._audio_capture_stream is None:
                # Set bounded capacity on AudioStream's internal queue to prevent unbounded memory growth
                # Use same capacity as our recorder queue (~30 seconds of frames)
                # Audio frames arrive at ~100fps, so 30 seconds = 3000 frames
                stream_capacity = max(1000, self.video_fps * 100)
                # Use Opus-compatible settings
                self._audio_capture_stream = AudioStream.from_track(
                    track=track,
                    sample_rate=48000,  # Opus typically uses 48kHz
                    num_channels=2,  # Stereo
                    capacity=stream_capacity,
                )
                self._audio_capture_task = asyncio.create_task(
                    self._capture_audio_frames()
                )
                logger.debug(f"Started audio capture from track {track.sid} with capacity={stream_capacity}")

    async def _capture_video_frames(self) -> None:
        """Capture video frames from the video stream."""
        if not self._video_capture_stream:
            logger.warning("Video capture stream is None, cannot capture frames")
            return

        logger.debug("Starting video frame capture")
        try:
            first_frame = True
            frame_count = 0
            async for frame_event in self._video_capture_stream:
                if not self._is_recording:
                    logger.debug(f"Recording stopped, breaking video capture loop. Captured {frame_count} frames")
                    break
                # Store first frame timestamp as reference for synchronization
                if first_frame and self._start_time:
                    # Convert start_time (seconds) to microseconds for consistency
                    self._recording_start_time_us = int(self._start_time * 1_000_000)
                    first_frame = False
                    logger.debug(f"Captured first video frame. Queue size before put: {self._video_queue.qsize()}")
                
                # Put frame in queue, waiting if necessary
                # The bounded queue size limits memory growth while allowing backpressure
                await self._video_queue.put(frame_event)
                frame_count += 1
                self._stats.video_frames_recorded += 1
                if frame_count % 100 == 0:
                    logger.debug(f"Captured {frame_count} video frames. Queue size: {self._video_queue.qsize()}")
        except Exception as e:
            logger.error(f"Error capturing video frames: {e}", exc_info=True)
        finally:
            if self._video_capture_stream:
                await self._video_capture_stream.aclose()
                self._video_capture_stream = None

    async def _capture_audio_frames(self) -> None:
        """Capture audio frames from the audio stream."""
        if not self._audio_capture_stream:
            logger.warning("Audio capture stream is None, cannot capture frames")
            return

        logger.debug("Starting audio frame capture")
        try:
            first_frame = True
            frame_count = 0
            async for frame_event in self._audio_capture_stream:
                if not self._is_recording:
                    logger.debug(f"Recording stopped, breaking audio capture loop. Captured {frame_count} frames")
                    break
                
                if first_frame:
                    logger.debug(f"Captured first audio frame. Queue size before put: {self._audio_queue.qsize()}")
                    first_frame = False
                
                # Put frame in queue, waiting if necessary
                # The bounded queue size limits memory growth while allowing backpressure
                await self._audio_queue.put(frame_event)
                frame_count += 1
                self._stats.audio_frames_recorded += 1
                if frame_count % 500 == 0:
                    logger.debug(f"Captured {frame_count} audio frames. Queue size: {self._audio_queue.qsize()}")
        except Exception as e:
            logger.error(f"Error capturing audio frames: {e}", exc_info=True)
        finally:
            if self._audio_capture_stream:
                await self._audio_capture_stream.aclose()
                self._audio_capture_stream = None

    async def _incremental_encoding_loop(self) -> None:
        """Incremental encoding loop that processes frames as they arrive."""
        import tempfile
        
        logger.info("Incremental encoding loop started")
        
        # Create temporary file for incremental encoding
        # This will be moved to the final location when stop_recording is called
        temp_fd, temp_path = tempfile.mkstemp(suffix=".webm", prefix="livekit_recording_")
        import os
        os.close(temp_fd)  # Close file descriptor, we'll open it with av.open
        self._output_file_path = temp_path
        logger.debug(f"Created temporary output file: {temp_path}")
        
        container_initialized = False
        container_init_lock = asyncio.Lock()
        
        try:
            
            # Process frames from queues as they arrive
            video_frame_count = 0
            audio_frame_count = 0
            logger.debug(f"Starting frame processing. Video queue size: {self._video_queue.qsize()}, Audio queue size: {self._audio_queue.qsize()}")
            
            # Use separate tasks to process video and audio frames concurrently
            # This allows proper waiting for frames without blocking
            async def process_video_frames():
                nonlocal video_frame_count, container_initialized
                frames_processed = False
                logger.debug("Video frame processing task started")
                while self._is_recording or not self._video_queue.empty():
                    try:
                        # Wait for frame with timeout to check recording status periodically
                        if self._is_recording:
                            try:
                                logger.debug(f"Waiting for video frame. Queue size: {self._video_queue.qsize()}, Recording: {self._is_recording}")
                                frame_event = await asyncio.wait_for(
                                    self._video_queue.get(), timeout=0.1
                                )
                                frames_processed = True
                                logger.debug(f"Received video frame {video_frame_count + 1}")
                            except asyncio.TimeoutError:
                                continue
                        else:
                            # When recording stopped, process remaining frames
                            # Wait a bit longer to ensure all frames are captured
                            queue_size = self._video_queue.qsize()
                            logger.debug(f"Recording stopped. Processing remaining video frames. Queue size: {queue_size}")
                            try:
                                frame_event = await asyncio.wait_for(
                                    self._video_queue.get(), timeout=1.0
                                )
                                frames_processed = True
                                logger.debug(f"Processed remaining video frame {video_frame_count + 1}")
                            except (asyncio.TimeoutError, asyncio.QueueEmpty):
                                logger.debug(f"Video queue empty. Processed {video_frame_count} frames total")
                                break
                        
                        async with container_init_lock:
                            if not container_initialized:
                                # Initialize container when first frame arrives
                                output_file = Path(self._output_file_path).resolve()
                                output_file.parent.mkdir(parents=True, exist_ok=True)
                                self._output_container = av.open(str(output_file), mode="w", format="webm")
                                container_initialized = True
                                self._container_was_initialized = True
                                logger.info("Initialized output container for incremental encoding")
                        
                        if not self._video_stream_initialized:
                            # Initialize video stream from first frame
                            frame = frame_event.frame
                            video_width = frame.width
                            video_height = frame.height
                            
                            # Store first frame timestamp as reference for PTS calculation
                            self._first_video_frame_time = frame_event.timestamp_us
                            
                            # Calculate optimal bitrate if auto_bitrate is enabled
                            actual_bitrate = self._calculate_bitrate(video_width, video_height)
                            
                            self._video_stream = self._output_container.add_stream(
                                self.video_codec,
                                rate=self.video_fps,
                            )
                            self._video_stream.width = video_width
                            self._video_stream.height = video_height
                            self._video_stream.pix_fmt = "yuv420p"
                            
                            # Build encoding options with quality settings
                            encoding_options = self._get_video_encoding_options(actual_bitrate)
                            self._video_stream.options = encoding_options
                            logger.info(f"Video encoding: {video_width}x{video_height} @ {actual_bitrate/1_000_000:.2f} Mbps, quality={self.video_quality}")
                            # Set time_base for video (1/1000 for milliseconds-based timing)
                            self._video_stream.time_base = Fraction(1, 1000)
                            self._video_stream_initialized = True
                            logger.info(f"Initialized video stream: {video_width}x{video_height}, first timestamp: {self._first_video_frame_time}us")
                        
                        # Encode video frame
                        await self._encode_video_frame_incremental(frame_event, video_frame_count)
                        video_frame_count += 1
                        
                        # Release frame reference immediately after encoding
                        frame_event.frame = None  # type: ignore
                        
                    except Exception as e:
                        logger.error(f"Error processing video frame: {e}", exc_info=True)
                        break
            
            async def process_audio_frames():
                nonlocal audio_frame_count, container_initialized
                frames_processed = False
                logger.debug("Audio frame processing task started")
                while self._is_recording or not self._audio_queue.empty():
                    try:
                        # Wait for frame with timeout to check recording status periodically
                        if self._is_recording:
                            try:
                                logger.debug(f"Waiting for audio frame. Queue size: {self._audio_queue.qsize()}, Recording: {self._is_recording}")
                                frame_event = await asyncio.wait_for(
                                    self._audio_queue.get(), timeout=0.1
                                )
                                frames_processed = True
                                logger.debug(f"Received audio frame {audio_frame_count + 1}")
                            except asyncio.TimeoutError:
                                continue
                        else:
                            # When recording stopped, process remaining frames
                            # Wait a bit longer to ensure all frames are captured
                            queue_size = self._audio_queue.qsize()
                            logger.debug(f"Recording stopped. Processing remaining audio frames. Queue size: {queue_size}")
                            try:
                                frame_event = await asyncio.wait_for(
                                    self._audio_queue.get(), timeout=1.0
                                )
                                frames_processed = True
                                logger.debug(f"Processed remaining audio frame {audio_frame_count + 1}")
                            except (asyncio.TimeoutError, asyncio.QueueEmpty):
                                logger.debug(f"Audio queue empty. Processed {audio_frame_count} frames total")
                                break
                        
                        async with container_init_lock:
                            if not container_initialized:
                                # Initialize container when first frame arrives
                                output_file = Path(self._output_file_path).resolve()
                                output_file.parent.mkdir(parents=True, exist_ok=True)
                                self._output_container = av.open(str(output_file), mode="w", format="webm")
                                container_initialized = True
                                self._container_was_initialized = True
                                logger.info("Initialized output container for incremental encoding")
                        
                        if not self._audio_stream_initialized:
                            # Initialize audio stream from first frame
                            frame = frame_event.frame
                            
                            self._audio_stream = self._output_container.add_stream(self.audio_codec)
                            self._audio_stream.rate = frame.sample_rate
                            # Set layout based on number of channels
                            if frame.num_channels == 1:
                                layout_str = "mono"
                            elif frame.num_channels == 2:
                                layout_str = "stereo"
                            else:
                                layout_str = f"{frame.num_channels}"
                            self._audio_stream.codec_context.layout = layout_str
                            self._audio_stream.options = {"bitrate": str(self.audio_bitrate)}
                            # Set time_base to match sample rate to avoid timestamp issues
                            self._audio_stream.time_base = Fraction(1, frame.sample_rate)
                            self._audio_stream_initialized = True
                            logger.info(f"Initialized audio stream: {frame.sample_rate}Hz, {frame.num_channels}ch")
                        
                        # Encode audio frame
                        await self._encode_audio_frame_incremental(frame_event)
                        audio_frame_count += 1
                        
                        # Release frame reference immediately after encoding
                        frame_event.frame = None  # type: ignore
                        
                    except Exception as e:
                        logger.error(f"Error processing audio frame: {e}", exc_info=True)
                        break
            
            # Process video and audio frames concurrently
            logger.debug("Starting concurrent frame processing tasks")
            await asyncio.gather(
                process_video_frames(),
                process_audio_frames(),
            )
            logger.debug("Frame processing tasks completed")
            
            # Flush encoders
            if self._video_stream:
                for packet in self._video_stream.encode():
                    if packet.dts is None:
                        continue
                    # Enforce monotonicity
                    if self._last_video_dts != -1 and packet.dts <= self._last_video_dts:
                        packet.dts = self._last_video_dts + 1
                        if packet.pts is not None and packet.pts < packet.dts:
                            packet.pts = packet.dts
                    self._last_video_dts = packet.dts
                    
                    # Ensure duration > 0 to prevent 0xc0000094 crash
                    if not packet.duration or packet.duration <= 0:
                        packet.duration = 1
                        
                    try:
                        self._output_container.mux(packet)
                    except Exception:
                        pass

            if self._audio_stream:
                for packet in self._audio_stream.encode():
                    if packet.dts is None:
                        continue
                    # Enforce monotonicity
                    if self._last_audio_dts != -1 and packet.dts <= self._last_audio_dts:
                        packet.dts = self._last_audio_dts + 1
                        if packet.pts is not None and packet.pts < packet.dts:
                            packet.pts = packet.dts
                    self._last_audio_dts = packet.dts
                    
                    # Ensure duration > 0 to prevent crash
                    if not packet.duration or packet.duration <= 0:
                        packet.duration = 1
                        
                    try:
                        self._output_container.mux(packet)
                    except Exception:
                        pass
            
            logger.info(f"Incremental encoding completed: {video_frame_count} video, {audio_frame_count} audio frames")
            
            # If no frames were processed, log a warning
            if video_frame_count == 0 and audio_frame_count == 0:
                logger.warning("No frames were processed during incremental encoding")
                if not container_initialized:
                    logger.warning("Container was never initialized - no frames arrived")
            
        except Exception as e:
            logger.error(f"Error in incremental encoding loop: {e}", exc_info=True)
            raise
        finally:
            if self._output_container:
                try:
                    # Flush the container before closing to ensure all data is written
                    # This is especially important for VP9 encoding
                    try:
                        self._output_container.flush()
                    except AttributeError:
                        # flush() might not be available in all PyAV versions
                        pass
                    self._output_container.close()
                    logger.info(f"Output container closed, temp file at: {self._output_file_path}")
                except Exception as e:
                    logger.error(f"Error closing output container: {e}", exc_info=True)
                self._output_container = None
            else:
                if self._output_file_path:
                    logger.debug(f"No container was initialized, temp file remains at: {self._output_file_path}")

    def _calculate_bitrate(self, width: int, height: int) -> int:
        """Calculate optimal bitrate based on resolution.
        
        Args:
            width: Video width in pixels.
            height: Video height in pixels.
            
        Returns:
            Optimal bitrate in bits per second.
        """
        if not self.auto_bitrate:
            return self.video_bitrate
        
        # Calculate pixels
        pixels = width * height
        
        # Bitrate recommendations based on resolution (bits per second per pixel * fps factor)
        # These are conservative estimates that work well for VP8/VP9
        # Adjust multipliers for different quality expectations
        
        # Base bitrate per megapixel at 30fps
        if pixels <= 640 * 480:  # VGA or smaller
            bitrate_per_megapixel = 2_000_000  # 2 Mbps per MP
        elif pixels <= 1280 * 720:  # 720p
            bitrate_per_megapixel = 3_000_000  # 3 Mbps per MP
        elif pixels <= 1920 * 1080:  # 1080p
            bitrate_per_megapixel = 5_000_000  # 5 Mbps per MP
        else:  # 1440p, 4K, etc.
            bitrate_per_megapixel = 8_000_000  # 8 Mbps per MP
        
        # Calculate base bitrate
        megapixels = pixels / 1_000_000
        calculated_bitrate = int(megapixels * bitrate_per_megapixel)
        
        # Apply quality multiplier
        quality_multipliers = {
            "low": 0.7,
            "medium": 1.0,
            "high": 1.5,
            "best": 2.0,
        }
        multiplier = quality_multipliers.get(self.video_quality, 1.0)
        calculated_bitrate = int(calculated_bitrate * multiplier)
        
        # Ensure minimum bitrate and use user's base bitrate as minimum
        min_bitrate = max(self.video_bitrate, 1_000_000)  # At least 1 Mbps
        return max(calculated_bitrate, min_bitrate)
    
    def _get_video_encoding_options(self, bitrate: int) -> dict[str, str]:
        """Get video encoding options based on codec and quality settings.
        
        Args:
            bitrate: Target bitrate in bits per second.
            
        Returns:
            Dictionary of encoding options for PyAV.
        """
        options: dict[str, str] = {
            "bitrate": str(bitrate),
        }
        
        if self.video_codec == "vp8":
            # VP8 quality/CPU tradeoff settings
            # cpu-used: 0-16, lower = better quality but slower
            cpu_used_map = {
                "low": "8",      # Fast encoding, lower quality
                "medium": "4",   # Balanced
                "high": "2",     # Slower encoding, better quality
                "best": "0",     # Slowest encoding, best quality
            }
            options["cpu-used"] = cpu_used_map.get(self.video_quality, "4")
            
            # Deadzone (noise sensitivity): 0-1000, lower = more sensitive to noise
            # Lower values preserve more detail but may introduce artifacts
            if self.video_quality in ("high", "best"):
                options["deadline"] = "goodquality"  # Better quality mode
                options["deadline_b"] = "600000"  # ~600ms per frame
            else:
                options["deadline"] = "realtime"  # Faster encoding
        
        elif self.video_codec == "vp9":
            # VP9 uses CRF (Constant Rate Factor) for quality-based encoding
            # CRF: 0-63, lower = better quality (0 = lossless, 31 = default, 63 = worst)
            crf_map = {
                "low": "45",     # Higher CRF = lower quality, smaller file
                "medium": "35",  # Default-like
                "high": "28",    # Better quality
                "best": "20",    # Very high quality
            }
            options["crf"] = crf_map.get(self.video_quality, "35")
            
            # CPU usage: 0-8, lower = better quality but slower
            cpu_used_map = {
                "low": "5",      # Faster encoding
                "medium": "3",   # Balanced
                "high": "1",     # Slower encoding, better quality
                "best": "0",     # Slowest encoding, best quality
            }
            options["cpu-used"] = cpu_used_map.get(self.video_quality, "3")
            
            # VP9 row-based multithreading (faster encoding)
            options["row-mt"] = "1"
            
            # For VP9, bitrate is used as max bitrate when CRF is set
            # The encoder will try to maintain quality while respecting bitrate limits
        
        return options

    def _encode_video_frame_incremental_sync(self, frame_event: VideoFrameEvent, frame_index: int) -> None:
        """Encode a single video frame incrementally (synchronous)."""
        if not self._video_stream or not self._output_container:
            return
        
        # Calculate PTS based on actual frame timestamps (not frame index)
        # This ensures correct duration regardless of actual frame rate
        if self._video_stream.time_base:
            time_base_denominator = self._video_stream.time_base.denominator
            time_base_numerator = self._video_stream.time_base.numerator
        else:
            time_base_denominator = 1000
            time_base_numerator = 1
        
        # Use actual timestamp from frame event to calculate PTS
        # timestamp_us is in microseconds, convert to PTS units
        if self._first_video_frame_time is not None:
            # Calculate time since first frame in seconds
            time_since_first_sec = (frame_event.timestamp_us - self._first_video_frame_time) / 1_000_000.0
            # Convert to PTS units: (time_sec * time_base_denominator) / time_base_numerator
            # Since time_base = 1/1000, PTS = time_ms = time_sec * 1000
            pts = int(time_since_first_sec * time_base_denominator / time_base_numerator)
        else:
            # Fallback to frame index if first timestamp not set (shouldn't happen)
            frame_interval = int(time_base_denominator / (self.video_fps * time_base_numerator))
            if frame_interval < 1:
                frame_interval = 1
            pts = frame_index * frame_interval
        
        # Convert and encode frame
        frame = frame_event.frame
        pyav_frame = self._convert_video_frame_to_pyav(frame, self._video_stream, pts)
        if pyav_frame and self._video_stream:
            for packet in self._video_stream.encode(pyav_frame):
                # Enforce monotonicity to prevent invalid argument error
                if packet.dts is None:
                    continue
                if self._last_video_dts != -1 and packet.dts <= self._last_video_dts:
                    packet.dts = self._last_video_dts + 1
                    if packet.pts is not None and packet.pts < packet.dts:
                        packet.pts = packet.dts
                self._last_video_dts = packet.dts
                
                # Ensure duration > 0 to prevent 0xc0000094 crash
                if not packet.duration or packet.duration <= 0:
                    packet.duration = 1
                    
                try:
                    self._output_container.mux(packet)
                except Exception as e:
                    logger.warning(f"Video mux failed: {e}")
        
        # Release frame reference immediately
        del pyav_frame
        
        # Periodic GC (every 50 frames)
        if frame_index > 0 and frame_index % 50 == 0:
            gc.collect()

    async def _encode_video_frame_incremental(self, frame_event: VideoFrameEvent, frame_index: int) -> None:
        """Encode a single video frame incrementally."""
        # Encoding is fast enough that we can do it directly in async context
        # PyAV encode operations are typically < 10ms per frame
        self._encode_video_frame_incremental_sync(frame_event, frame_index)

    def _encode_audio_frame_incremental_sync(self, frame_event: AudioFrameEvent) -> None:
        """Encode a single audio frame incrementally (synchronous)."""
        if not self._audio_stream or not self._output_container:
            return
        
        frame = frame_event.frame
        sample_rate = frame.sample_rate
        samples_per_channel = frame.samples_per_channel
        
        # Calculate audio PTS based on cumulative samples
        # Use time_base of audio stream (1/sample_rate) to avoid timestamp issues
        # PTS = cumulative_samples (since time_base is 1/sample_rate)
        audio_pts = self._cumulative_audio_samples
        
        # Convert and encode frame
        pyav_frame = self._convert_audio_frame_to_pyav(frame, self._audio_stream, audio_pts)
        if pyav_frame and self._audio_stream:
            for packet in self._audio_stream.encode(pyav_frame):
                # Enforce monotonicity
                if packet.dts is None:
                    continue
                if self._last_audio_dts != -1 and packet.dts <= self._last_audio_dts:
                    packet.dts = self._last_audio_dts + 1
                    if packet.pts is not None and packet.pts < packet.dts:
                        packet.pts = packet.dts
                self._last_audio_dts = packet.dts

                # Ensure duration > 0
                if not packet.duration or packet.duration <= 0:
                    packet.duration = 1

                try:
                    self._output_container.mux(packet)
                except Exception as e:
                    logger.warning(f"Audio mux failed: {e}")
            
            # Update cumulative samples for next frame
            self._cumulative_audio_samples += samples_per_channel
        
        # Release frame reference immediately
        del pyav_frame
        
        # Periodic GC (every ~4 seconds of audio)
        if self._cumulative_audio_samples % (sample_rate * 4) == 0:
            gc.collect()

    async def _encode_audio_frame_incremental(self, frame_event: AudioFrameEvent) -> None:
        """Encode a single audio frame incrementally."""
        # Encoding is fast enough that we can do it directly in async context
        self._encode_audio_frame_incremental_sync(frame_event)

    async def stop_recording(self, output_path: str) -> str:
        """Stop recording and save to a WebM file.

        Args:
            output_path: Path where the WebM file should be saved.

        Returns:
            The absolute path to the saved file.

        Raises:
            RecordingError: If recording is not in progress or encoding fails.
        """
        async with self._lock:
            if not self._is_recording:
                raise RecordingError("Recording is not in progress")

            self._is_recording = False

            # Update statistics
            if self._start_time:
                self._stats.recording_duration_seconds = (
                    time.time() - self._start_time
                )

            logger.info("Stopping recording and encoding to WebM...")

            # Stop capturing tasks
            if self._video_capture_task:
                self._video_capture_task.cancel()
                try:
                    await self._video_capture_task
                except asyncio.CancelledError:
                    pass

            if self._audio_capture_task:
                self._audio_capture_task.cancel()
                try:
                    await self._audio_capture_task
                except asyncio.CancelledError:
                    pass

            # Close streams
            if self._video_capture_stream:
                try:
                    await self._video_capture_stream.aclose()
                except asyncio.CancelledError:
                    pass
                self._video_capture_stream = None

            if self._audio_capture_stream:
                try:
                    await self._audio_capture_stream.aclose()
                except asyncio.CancelledError:
                    pass
                self._audio_capture_stream = None

            # Unsubscribe from tracks to stop network activity
            # This must be done before waiting for encoding to complete
            # to ensure the server stops sending data immediately
            if self._participant:
                await self._unsubscribe_from_participant_tracks(self._participant)

            # Wait for encoding task to complete and flush/close
            if self._encoding_task:
                try:
                    await self._encoding_task
                except Exception as e:
                    logger.error(f"Error during incremental encoding: {e}", exc_info=True)
                    raise RecordingError(f"Failed to complete encoding: {e}") from e
            
            # Move temporary file to final location
            import shutil
            import os
            import time as time_module
            
            # Give a small delay to ensure file system writes are complete
            # This is especially important for large files or slow storage
            await asyncio.sleep(0.1)
            
            if self._output_file_path and os.path.exists(self._output_file_path):
                # Check if container was initialized (file should have content)
                file_size = os.path.getsize(self._output_file_path)
                logger.debug(f"Temp file exists: {self._output_file_path}, size: {file_size} bytes, container_initialized: {self._container_was_initialized}")
                
                if file_size > 0 or self._container_was_initialized:
                    # Double-check file size after a brief delay (for file system sync)
                    time_module.sleep(0.1)
                    final_size = os.path.getsize(self._output_file_path)
                    logger.info(f"Moving encoded file: {self._output_file_path} ({final_size} bytes) -> {output_path}")
                    
                    output_file_obj = Path(output_path).resolve()
                    output_file_obj.parent.mkdir(parents=True, exist_ok=True)
                    shutil.move(self._output_file_path, str(output_file_obj))
                    output_file = str(output_file_obj)
                    
                    # Verify the moved file exists and has content
                    if os.path.exists(output_file):
                        moved_size = os.path.getsize(output_file)
                        logger.info(f"Moved encoded file from {self._output_file_path} to {output_file} (size: {moved_size} bytes)")
                    else:
                        logger.error(f"File move failed: {output_file} does not exist after move")
                        raise RecordingError(f"Failed to move encoded file to {output_file}")
                else:
                    # Container was never initialized (no frames arrived)
                    logger.warning(f"No frames were encoded. Container was never initialized. Temp file size: {file_size} bytes")
                    # Remove empty temp file
                    try:
                        os.unlink(self._output_file_path)
                    except Exception:
                        pass
                    # Create empty output file or raise error
                    output_file_obj = Path(output_path).resolve()
                    output_file_obj.parent.mkdir(parents=True, exist_ok=True)
                    output_file_obj.touch()
                    output_file = str(output_file_obj)
            else:
                # No temp file was created (encoding task might have failed)
                logger.error(f"No output file was created. _output_file_path: {self._output_file_path}, container_initialized: {self._container_was_initialized}")
                if self._output_file_path:
                    logger.error(f"Temp file path was set but doesn't exist: {self._output_file_path}")
                    # Check if temp file exists in a different location
                    import glob
                    temp_pattern = os.path.join(os.path.dirname(self._output_file_path) if self._output_file_path else "/tmp", "livekit_recording_*.webm")
                    found_files = glob.glob(temp_pattern)
                    if found_files:
                        logger.info(f"Found potential temp files: {found_files}")
                        # Try using the most recent one
                        latest_file = max(found_files, key=os.path.getmtime)
                        file_size = os.path.getsize(latest_file)
                        if file_size > 0:
                            logger.info(f"Using found temp file: {latest_file} ({file_size} bytes)")
                            output_file_obj = Path(output_path).resolve()
                            output_file_obj.parent.mkdir(parents=True, exist_ok=True)
                            shutil.move(latest_file, str(output_file_obj))
                            output_file = str(output_file_obj)
                            logger.info(f"Moved found temp file to {output_file} (size: {file_size} bytes)")
                            return output_file
                
                output_file_obj = Path(output_path).resolve()
                output_file_obj.parent.mkdir(parents=True, exist_ok=True)
                output_file_obj.touch()
                output_file = str(output_file_obj)
                logger.warning("Created empty output file as fallback")

            # Clean up all references to allow proper resource release
            self._participant_identity = None
            self._participant = None
            self._subscribed_track_sids.clear()
            # Clear room reference to allow room to be disconnected
            self.room = None  # type: ignore

            logger.info(f"Recording saved to {output_file}")
            return output_file

    async def _encode_to_webm(self, output_path: str) -> str:
        """Encode captured frames to a WebM file incrementally.
        
        This method processes frames as they're dequeued rather than collecting
        all frames first, significantly reducing memory usage for long recordings.
        """
        output_file = Path(output_path).resolve()
        output_file.parent.mkdir(parents=True, exist_ok=True)

        # Check if we have any frames to encode
        if self._video_queue.empty() and self._audio_queue.empty():
            logger.warning(
                "No frames captured during recording. Creating empty file will be skipped."
            )
            # Create an empty file but mark it as such
            output_file.touch()
            return str(output_file)

        # Process frames incrementally - collect them in small batches and sort
        # This allows us to process frames in order while limiting memory usage
        # We'll collect frames in batches, sort each batch, then encode incrementally
        video_frames_batch: list[VideoFrameEvent] = []
        audio_frames_batch: list[AudioFrameEvent] = []
        
        # Collect remaining frames from queues (queues are bounded so this is limited)
        while True:
            try:
                frame_event = self._video_queue.get_nowait()
                video_frames_batch.append(frame_event)
            except asyncio.QueueEmpty:
                break

        while True:
            try:
                frame_event = self._audio_queue.get_nowait()
                audio_frames_batch.append(frame_event)
            except asyncio.QueueEmpty:
                break

        # Sort frames by timestamp to ensure proper synchronization order
        # Since queues are bounded (max ~5 seconds of frames), sorting is still efficient
        video_frames_batch.sort(key=lambda e: e.timestamp_us)
        # Audio frames don't have timestamps, but we'll process them in order

        # Run encoding in a thread pool to avoid blocking
        loop = asyncio.get_event_loop()
        absolute_path = await loop.run_in_executor(
            None,
            self._encode_to_webm_sync,
            str(output_file),
            video_frames_batch,
            audio_frames_batch,
        )

        # Clear frames from memory after encoding to prevent memory leaks
        del video_frames_batch
        del audio_frames_batch

        # Get file size
        output_file_obj = Path(absolute_path)
        if output_file_obj.exists():
            self._stats.output_file_size_bytes = output_file_obj.stat().st_size

        return absolute_path

    def _encode_to_webm_sync(
        self,
        output_path: str,
        video_frames: list[VideoFrameEvent],
        audio_frames: list[AudioFrameEvent],
    ) -> str:
        """Synchronously encode frames to WebM (runs in executor)."""
        logger.info(f"Starting WebM encoding: {len(video_frames)} video frames, {len(audio_frames)} audio frames")
        # Create output container
        output_container = av.open(output_path, mode="w", format="webm")

        try:
            video_stream: Optional[av.VideoStream] = None
            audio_stream: Optional[av.AudioStream] = None

            # Determine video properties from first frame
            if video_frames:
                first_video_frame = video_frames[0].frame
                video_width = first_video_frame.width
                video_height = first_video_frame.height

                # Add video stream
                video_stream = output_container.add_stream(
                    self.video_codec,
                    rate=self.video_fps,
                )
                video_stream.width = video_width
                video_stream.height = video_height
                video_stream.pix_fmt = "yuv420p"
                # Calculate optimal bitrate if auto_bitrate is enabled
                actual_bitrate = self._calculate_bitrate(video_width, video_height)
                # Build encoding options with quality settings
                video_stream.options = self._get_video_encoding_options(actual_bitrate)

            # Add audio stream
            if audio_frames:
                first_audio_frame = audio_frames[0].frame
                audio_stream = output_container.add_stream(self.audio_codec)
                audio_stream.rate = first_audio_frame.sample_rate
                # Set layout (channels is read-only on stream and codec_context)
                # Determine layout string based on number of channels
                if first_audio_frame.num_channels == 1:
                    layout_str = "mono"
                elif first_audio_frame.num_channels == 2:
                    layout_str = "stereo"
                else:
                    layout_str = f"{first_audio_frame.num_channels}"
                audio_stream.codec_context.layout = layout_str
                audio_stream.options = {"bitrate": str(self.audio_bitrate)}

            # Encode video frames using frame-based PTS calculation
            # Calculate spacing based on actual recording duration to ensure correct playback speed
            if video_stream:
                total_video_frames = len(video_frames)
                logger.debug(f"Encoding {total_video_frames} video frames...")
                
                # Calculate frame interval in time_base units
                if video_stream.time_base:
                    time_base_denominator = video_stream.time_base.denominator
                    time_base_numerator = video_stream.time_base.numerator
                else:
                    time_base_denominator = 1000
                    time_base_numerator = 1
                
                # Calculate actual frame rate from recording duration
                # This ensures video plays at correct speed over the full duration
                recording_duration_sec = self._stats.recording_duration_seconds
                if recording_duration_sec > 0 and total_video_frames > 0:
                    # Calculate actual frame rate from captured frames
                    actual_fps = total_video_frames / recording_duration_sec
                    logger.debug(f"Recording duration: {recording_duration_sec:.2f}s, frames: {total_video_frames}, actual FPS: {actual_fps:.2f}")
                else:
                    # Fallback to target FPS
                    actual_fps = self.video_fps
                
                # Calculate frame interval based on actual frame rate
                # This ensures frames are evenly spaced over the recording duration
                # frame_interval = time_base_denominator / (actual_fps * time_base_numerator)
                frame_interval = int(time_base_denominator / (actual_fps * time_base_numerator))
                if frame_interval < 1:
                    frame_interval = 1
                
                logger.debug(f"Frame interval: {frame_interval} PTS units ({frame_interval/1000:.3f} ms per frame at {actual_fps:.2f} fps)")
                
                # Calculate PTS for each frame based on frame index
                # This ensures frames are spaced correctly at the target frame rate
                for idx, frame_event in enumerate(video_frames):
                    if idx > 0 and idx % 100 == 0:
                        logger.debug(f"Encoding video frame {idx}/{len(video_frames)}...")
                    frame = frame_event.frame
                    
                    # Calculate PTS based on frame index
                    # PTS = frame_index * frame_interval
                    # Frame 0: PTS = 0 (starts immediately)
                    # Frame 1: PTS = frame_interval (~33ms at 30fps)
                    # Frame 30: PTS = 30 * frame_interval (~1 second at 30fps)
                    pts = idx * frame_interval
                    
                    # Convert VideoFrame to PyAV frame
                    pyav_frame = self._convert_video_frame_to_pyav(
                        frame, video_stream, pts
                    )
                    if pyav_frame and video_stream:
                        for packet in video_stream.encode(pyav_frame):
                            output_container.mux(packet)
                    
                    # Explicitly release frame references to help GC
                    # This is critical for memory efficiency as frames can be large (~3-4MB each)
                    del pyav_frame
                    # Also release the frame reference from the event
                    # This helps Python GC free the underlying frame data sooner
                    frame_event.frame = None  # type: ignore
                    
                    # Trigger GC periodically to free memory (every 50 frames)
                    # This helps prevent memory accumulation during encoding
                    if idx > 0 and idx % 50 == 0:
                        gc.collect()

            # Flush video encoder
            if video_stream:
                for packet in video_stream.encode():
                    output_container.mux(packet)

            # Encode audio frames synchronized with video timestamps
            # Calculate audio PTS based on cumulative samples, aligned with video timing
            # Use the same time_base as video (1/1000 milliseconds) for consistency
            if audio_stream:
                sample_rate = audio_frames[0].frame.sample_rate if audio_frames else 48000
                
                # Calculate audio PTS in milliseconds (same time_base as video: 1/1000)
                # Audio PTS should match the video timing based on sample count
                cumulative_samples = 0
                
                for frame_event in audio_frames:
                    frame = frame_event.frame
                    samples_per_channel = frame.samples_per_channel
                    
                    # Calculate PTS in milliseconds based on cumulative samples
                    # This naturally aligns with video timing since both start from 0
                    # PTS_ms = (cumulative_samples / sample_rate) * 1000
                    audio_pts_ms = int((cumulative_samples / sample_rate) * 1000)
                    
                    # Convert AudioFrame to PyAV frame
                    pyav_frame = self._convert_audio_frame_to_pyav(
                        frame, audio_stream, audio_pts_ms
                    )
                    if pyav_frame and audio_stream:
                        for packet in audio_stream.encode(pyav_frame):
                            output_container.mux(packet)
                        # Increment by samples per channel for next frame
                        cumulative_samples += samples_per_channel
                    
                    # Explicitly release frame references to help GC
                    del pyav_frame
                    # Release the frame reference from the event to help GC free audio data
                    frame_event.frame = None  # type: ignore
                    
                    # Trigger GC periodically (every 200 audio frames since they're smaller)
                    if cumulative_samples % (sample_rate * 4) == 0:  # Every ~4 seconds of audio
                        gc.collect()

            # Flush audio encoder
            if audio_stream:
                for packet in audio_stream.encode():
                    output_container.mux(packet)

        finally:
            output_container.close()
            # Explicitly release all frame references to help GC
            # This is critical for memory efficiency - frames can be several MB each
            for frame_event in video_frames:
                if hasattr(frame_event, 'frame'):
                    frame_event.frame = None  # type: ignore
            for frame_event in audio_frames:
                if hasattr(frame_event, 'frame'):
                    frame_event.frame = None  # type: ignore
            video_frames.clear()
            audio_frames.clear()
            
            # Force garbage collection after encoding to free memory
            # This is important because frames can be large and GC might not run immediately
            gc.collect()

        return output_path

    def _convert_video_frame_to_pyav(
        self,
        frame: VideoFrame,
        stream: Optional[av.VideoStream],
        pts: int,
    ) -> Optional[av.VideoFrame]:
        """Convert a VideoFrame to a PyAV VideoFrame.
        
        Note: This creates a copy of frame data. For memory efficiency, frame references
        should be released after encoding (handled in calling code).
        """
        if not stream:
            return None

        try:
            # Convert frame to I420 format if needed
            # This creates a new VideoFrame - the old one will be GC'd after use
            converted_frame = None
            if frame.type != VideoBufferType.I420:
                converted_frame = frame.convert(VideoBufferType.I420)
                frame = converted_frame

            # Get I420 planes as memoryviews (these include stride padding)
            y_plane = frame.get_plane(0)
            u_plane = frame.get_plane(1)
            v_plane = frame.get_plane(2)

            if not y_plane or not u_plane or not v_plane:
                return None

            # Create PyAV frame directly and update planes
            # PyAV doesn't support from_ndarray with different-sized planes for YUV420P
            # So we create the frame and update planes manually
            pyav_frame = av.VideoFrame(frame.width, frame.height, "yuv420p")
            
            # Copy plane data accounting for stride differences
            # The memoryview from get_plane() includes stride padding
            # We need to copy the actual image data, skipping stride padding if needed
            
            # Calculate strides
            y_stride = len(y_plane) // frame.height
            pyav_y_stride = pyav_frame.planes[0].line_size
            chroma_height = frame.height // 2
            u_stride = len(u_plane) // chroma_height if chroma_height > 0 else 0
            v_stride = len(v_plane) // chroma_height if chroma_height > 0 else 0
            pyav_u_stride = pyav_frame.planes[1].line_size
            pyav_v_stride = pyav_frame.planes[2].line_size
            
            # Copy Y plane row by row, handling stride differences
            # Use memoryview for direct access without intermediate numpy copies where possible
            pyav_y_data = bytearray(pyav_frame.planes[0].buffer_size)
            y_mv = memoryview(y_plane)
            for row in range(frame.height):
                src_start = row * y_stride
                dst_start = row * pyav_y_stride
                copy_len = min(frame.width, min(y_stride, pyav_y_stride))
                # Direct slice assignment avoids tobytes() copy
                pyav_y_data[dst_start:dst_start + copy_len] = y_mv[src_start:src_start + copy_len]
            # Use memoryview for direct update if possible, otherwise use bytes()
            # PyAV's update() accepts bytes/memoryview, so we can pass the bytearray directly as memoryview
            pyav_frame.planes[0].update(memoryview(pyav_y_data))
            # Release intermediate buffers immediately
            del pyav_y_data, y_mv
            
            # Copy U plane row by row
            chroma_width = frame.width // 2
            pyav_u_data = bytearray(pyav_frame.planes[1].buffer_size)
            u_mv = memoryview(u_plane)
            for row in range(chroma_height):
                src_start = row * u_stride
                dst_start = row * pyav_u_stride
                copy_len = min(chroma_width, min(u_stride, pyav_u_stride))
                pyav_u_data[dst_start:dst_start + copy_len] = u_mv[src_start:src_start + copy_len]
            # Use memoryview instead of bytes() to avoid extra copy
            pyav_frame.planes[1].update(memoryview(pyav_u_data))
            # Release intermediate buffers immediately
            del pyav_u_data, u_mv
            
            # Copy V plane row by row
            pyav_v_data = bytearray(pyav_frame.planes[2].buffer_size)
            v_mv = memoryview(v_plane)
            for row in range(chroma_height):
                src_start = row * v_stride
                dst_start = row * pyav_v_stride
                copy_len = min(chroma_width, min(v_stride, pyav_v_stride))
                pyav_v_data[dst_start:dst_start + copy_len] = v_mv[src_start:src_start + copy_len]
            # Use memoryview instead of bytes() to avoid extra copy
            pyav_frame.planes[2].update(memoryview(pyav_v_data))
            # Release intermediate buffers immediately
            del pyav_v_data, v_mv
            
            pyav_frame.pts = pts
            # Set time_base only if stream has one and it's valid
            if stream and hasattr(stream, 'time_base') and stream.time_base is not None:
                try:
                    pyav_frame.time_base = stream.time_base
                except (AttributeError, ValueError, TypeError) as e:
                    logger.warning(f"Could not set time_base: {e}")
                    # Continue without time_base - PyAV may infer it

            # Release converted frame reference if we created one
            # This helps GC free the conversion buffer sooner
            if converted_frame is not None:
                del converted_frame
            
            return pyav_frame
        except Exception as e:
            logger.error(f"Error converting video frame: {e}", exc_info=True)
            return None

    def _convert_audio_frame_to_pyav(
        self,
        frame: AudioFrame,
        stream: Optional[av.AudioStream],
        pts: int,
    ) -> Optional[av.AudioFrame]:
        """Convert an AudioFrame to a PyAV AudioFrame."""
        if not stream:
            return None

        try:
            # Get audio data as numpy array
            import numpy as np

            audio_data = np.frombuffer(frame.data, dtype=np.int16)

            # PyAV expects "packed" format: 1D array with shape (1, samples*channels)
            # AudioFrame.data is already interleaved (channels interleaved)
            # Reshape to (1, samples*channels) for PyAV's packed format
            audio_packed = audio_data.reshape(1, -1)

            # Determine layout string
            if frame.num_channels == 1:
                layout = "mono"
            elif frame.num_channels == 2:
                layout = "stereo"
            else:
                layout = f"{frame.num_channels}"

            # Create PyAV frame with packed format (1, samples*channels)
            pyav_frame = av.AudioFrame.from_ndarray(
                audio_packed,
                format="s16",
                layout=layout,
            )
            pyav_frame.sample_rate = frame.sample_rate
            pyav_frame.pts = pts
            # Set time_base only if stream has one (it may be None initially)
            if stream and stream.time_base:
                pyav_frame.time_base = stream.time_base

            return pyav_frame
        except Exception as e:
            logger.error(f"Error converting audio frame: {e}", exc_info=True)
            return None

    def get_stats(self) -> RecordingStats:
        """Get statistics about the current or completed recording.

        Returns:
            RecordingStats object with recording statistics.
        """
        if self._start_time and self._is_recording:
            self._stats.recording_duration_seconds = (
                time.time() - self._start_time
            )
        return RecordingStats(
            video_frames_recorded=self._stats.video_frames_recorded,
            audio_frames_recorded=self._stats.audio_frames_recorded,
            recording_duration_seconds=self._stats.recording_duration_seconds,
            output_file_size_bytes=self._stats.output_file_size_bytes,
        )

    @property
    def is_recording(self) -> bool:
        """Check if recording is currently in progress."""
        return self._is_recording



#!/usr/bin/env python3
"""Memory monitoring test for ParticipantRecorder.

This test runs a 20-minute recording session and monitors memory usage every 60 seconds.
It helps identify memory leaks or memory growth issues during long recordings.
"""

import asyncio
import os
import sys
import time
import tempfile
import logging
import json
import subprocess
import gc
import tracemalloc
from pathlib import Path
from typing import Dict, List, Any, Optional
from dataclasses import dataclass, asdict

# Try to import psutil, fall back to system commands if not available
try:
    import psutil
    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

# Add parent directory to path for direct imports
sys.path.insert(0, str(Path(__file__).parent.parent))
sys.path.insert(0, str(Path(__file__).parent.parent.parent / "livekit-protocol"))
sys.path.insert(0, str(Path(__file__).parent.parent.parent / "livekit-api"))

from livekit import rtc
from livekit import api as livekit_api
from livekit.rtc.recorder import (
    ParticipantRecorder,
    ParticipantNotFoundError,
    RecordingError,
)

# Import integration test utilities
try:
    from .test_recorder_integration import wait_for_participant, wait_for_tracks
    from .pumpfun_service import PumpFunService
    HAS_PUMPFUN = True
except ImportError:
    try:
        from test_recorder_integration import wait_for_participant, wait_for_tracks
        from pumpfun_service import PumpFunService
        HAS_PUMPFUN = True
    except ImportError:
        HAS_PUMPFUN = False

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger(__name__)


@dataclass
class MemorySnapshot:
    """Memory usage snapshot at a specific time."""
    timestamp: float
    elapsed_seconds: float
    rss_mb: float  # Resident Set Size in MB
    vms_mb: float  # Virtual Memory Size in MB
    percent: float  # Percentage of system memory
    available_mb: float  # Available system memory in MB
    video_frames: int
    audio_frames: int
    recording_duration: float
    # Profiling data
    tracemalloc_current_mb: Optional[float] = None
    tracemalloc_peak_mb: Optional[float] = None
    top_allocations: Optional[List[Dict[str, Any]]] = None
    object_counts: Optional[Dict[str, int]] = None


def get_memory_usage_system(pid: int) -> Dict[str, float]:
    """Get memory usage using system commands (fallback when psutil not available)."""
    try:
        # Use ps command on macOS/Linux
        result = subprocess.run(
            ["ps", "-o", "rss,vsz,%mem", "-p", str(pid)],
            capture_output=True,
            text=True,
            check=True,
        )
        lines = result.stdout.strip().split("\n")
        if len(lines) < 2:
            return {"rss_mb": 0.0, "vms_mb": 0.0, "percent": 0.0, "available_mb": 0.0}
        
        # Parse output: RSS VSZ %MEM
        parts = lines[1].split()
        if len(parts) >= 3:
            rss_kb = float(parts[0])
            vms_kb = float(parts[1])
            percent = float(parts[2])
            
            # Get system memory using vm_stat or sysctl
            try:
                if sys.platform == "darwin":  # macOS
                    result = subprocess.run(
                        ["sysctl", "hw.memsize"],
                        capture_output=True,
                        text=True,
                        check=True,
                    )
                    total_mem = int(result.stdout.split(":")[1].strip())
                    # Rough estimate of available memory (this is simplified)
                    available_mb = total_mem / 1024 / 1024 * 0.5  # Assume 50% available
                else:  # Linux
                    with open("/proc/meminfo", "r") as f:
                        meminfo = f.read()
                        for line in meminfo.split("\n"):
                            if "MemAvailable:" in line:
                                available_kb = int(line.split()[1])
                                available_mb = available_kb / 1024
                                break
                        else:
                            available_mb = 0
            except:
                available_mb = 0
            
            return {
                "rss_mb": rss_kb / 1024,  # Convert KB to MB
                "vms_mb": vms_kb / 1024,  # Convert KB to MB
                "percent": percent,
                "available_mb": available_mb,
            }
    except Exception as e:
        logger.warning(f"Error getting memory usage via system commands: {e}")
    
    return {"rss_mb": 0.0, "vms_mb": 0.0, "percent": 0.0, "available_mb": 0.0}


def get_memory_usage(process_or_pid) -> Dict[str, float]:
    """Get current memory usage for a process."""
    if HAS_PSUTIL:
        try:
            process = process_or_pid if hasattr(process_or_pid, "memory_info") else psutil.Process(process_or_pid)
            memory_info = process.memory_info()
            memory_percent = process.memory_percent()
            
            # Get system memory
            system_memory = psutil.virtual_memory()
            
            return {
                "rss_mb": memory_info.rss / 1024 / 1024,  # MB
                "vms_mb": memory_info.vms / 1024 / 1024,  # MB
                "percent": memory_percent,
                "available_mb": system_memory.available / 1024 / 1024,  # MB
            }
        except Exception as e:
            logger.warning(f"Error getting memory usage via psutil: {e}")
            # Fall back to system commands
            pid = process_or_pid if isinstance(process_or_pid, int) else process_or_pid.pid
            return get_memory_usage_system(pid)
    else:
        # Use system commands
        pid = process_or_pid if isinstance(process_or_pid, int) else process_or_pid.pid
        return get_memory_usage_system(pid)


def get_tracemalloc_stats() -> Dict[str, Any]:
    """Get tracemalloc statistics if tracing is active."""
    if not tracemalloc.is_tracing():
        return {}
    
    try:
        current, peak = tracemalloc.get_traced_memory()
        current_mb = current / 1024 / 1024
        peak_mb = peak / 1024 / 1024
        
        # Get top 10 allocations by size
        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics('lineno')
        
        top_allocations = []
        for index, stat in enumerate(top_stats[:10], 1):
            top_allocations.append({
                "rank": index,
                "size_mb": stat.size / 1024 / 1024,
                "count": stat.count,
                "filename": stat.traceback[0].filename if stat.traceback else "unknown",
                "lineno": stat.traceback[0].lineno if stat.traceback else 0,
            })
        
        return {
            "current_mb": current_mb,
            "peak_mb": peak_mb,
            "top_allocations": top_allocations,
        }
    except Exception as e:
        logger.warning(f"Error getting tracemalloc stats: {e}")
        return {}


def get_object_counts() -> Dict[str, int]:
    """Count objects by type in memory."""
    try:
        gc.collect()  # Force collection before counting
        objects = gc.get_objects()
        
        type_counts: Dict[str, int] = {}
        for obj in objects:
            obj_type = type(obj).__name__
            type_counts[obj_type] = type_counts.get(obj_type, 0) + 1
        
        # Sort by count and return top 20
        sorted_counts = sorted(type_counts.items(), key=lambda x: x[1], reverse=True)[:20]
        return dict(sorted_counts)
    except Exception as e:
        logger.warning(f"Error counting objects: {e}")
        return {}


async def monitor_memory(
    process_or_pid: Any,
    recorder: ParticipantRecorder,
    start_time: float,
    duration_seconds: float,
    interval_seconds: float = 60.0,
    snapshots: List[MemorySnapshot] = None,
    enable_profiling: bool = True,
) -> List[MemorySnapshot]:
    """Monitor memory usage at regular intervals during recording."""
    if snapshots is None:
        snapshots = []
    
    # Take initial snapshot at start
    memory_info = get_memory_usage(process_or_pid)
    stats = recorder.get_stats()
    
    # Get profiling data if enabled
    tracemalloc_stats = get_tracemalloc_stats() if enable_profiling else {}
    object_counts = get_object_counts() if enable_profiling else {}
    
    initial_snapshot = MemorySnapshot(
        timestamp=start_time,
        elapsed_seconds=0.0,
        rss_mb=memory_info["rss_mb"],
        vms_mb=memory_info["vms_mb"],
        percent=memory_info["percent"],
        available_mb=memory_info["available_mb"],
        video_frames=stats.video_frames_recorded,
        audio_frames=stats.audio_frames_recorded,
        recording_duration=stats.recording_duration_seconds,
        tracemalloc_current_mb=tracemalloc_stats.get("current_mb"),
        tracemalloc_peak_mb=tracemalloc_stats.get("peak_mb"),
        top_allocations=tracemalloc_stats.get("top_allocations"),
        object_counts=object_counts,
    )
    snapshots.append(initial_snapshot)
    
    log_msg = (
        f"[0.0s] Initial Memory: RSS={initial_snapshot.rss_mb:.1f}MB, "
        f"VMS={initial_snapshot.vms_mb:.1f}MB, Percent={initial_snapshot.percent:.1f}%"
    )
    if enable_profiling and tracemalloc_stats:
        log_msg += f", Traced: {tracemalloc_stats.get('current_mb', 0):.1f}MB"
    logger.info(log_msg)
    
    elapsed = 0.0
    check_count = 0
    
    while elapsed < duration_seconds:
        await asyncio.sleep(interval_seconds)
        elapsed = time.time() - start_time
        
        if elapsed >= duration_seconds:
            break
        
        check_count += 1
        memory_info = get_memory_usage(process_or_pid)
        stats = recorder.get_stats()
        
        # Get profiling data if enabled
        tracemalloc_stats = get_tracemalloc_stats() if enable_profiling else {}
        object_counts = get_object_counts() if enable_profiling else {}
        
        snapshot = MemorySnapshot(
            timestamp=time.time(),
            elapsed_seconds=elapsed,
            rss_mb=memory_info["rss_mb"],
            vms_mb=memory_info["vms_mb"],
            percent=memory_info["percent"],
            available_mb=memory_info["available_mb"],
            video_frames=stats.video_frames_recorded,
            audio_frames=stats.audio_frames_recorded,
            recording_duration=stats.recording_duration_seconds,
            tracemalloc_current_mb=tracemalloc_stats.get("current_mb"),
            tracemalloc_peak_mb=tracemalloc_stats.get("peak_mb"),
            top_allocations=tracemalloc_stats.get("top_allocations"),
            object_counts=object_counts,
        )
        snapshots.append(snapshot)
        
        log_msg = (
            f"[{elapsed:.1f}s] Memory: RSS={snapshot.rss_mb:.1f}MB, "
            f"VMS={snapshot.vms_mb:.1f}MB, Percent={snapshot.percent:.1f}%, "
            f"Frames: {snapshot.video_frames} video, {snapshot.audio_frames} audio"
        )
        if enable_profiling and tracemalloc_stats:
            log_msg += f", Traced: {tracemalloc_stats.get('current_mb', 0):.1f}MB"
        logger.info(log_msg)
        
        # Log top allocations periodically (every 5 minutes)
        if enable_profiling and check_count % 5 == 0 and tracemalloc_stats.get("top_allocations"):
            logger.info("  Top 5 allocations:")
            for alloc in tracemalloc_stats["top_allocations"][:5]:
                logger.info(
                    f"    {alloc['filename']}:{alloc['lineno']} - "
                    f"{alloc['size_mb']:.2f}MB ({alloc['count']} allocations)"
                )
        
        # Log top object types periodically (every 5 minutes)
        if enable_profiling and check_count % 5 == 0 and object_counts:
            logger.info("  Top 5 object types:")
            for obj_type, count in list(object_counts.items())[:5]:
                logger.info(f"    {obj_type}: {count:,} objects")
    
    return snapshots


def analyze_memory_usage(snapshots: List[MemorySnapshot]) -> Dict[str, Any]:
    """Analyze memory usage trends from snapshots."""
    if not snapshots:
        return {"error": "No snapshots available"}
    
    rss_values = [s.rss_mb for s in snapshots]
    vms_values = [s.vms_mb for s in snapshots]
    percent_values = [s.percent for s in snapshots]
    
    # Calculate trends
    initial_rss = rss_values[0] if rss_values else 0
    final_rss = rss_values[-1] if rss_values else 0
    rss_growth = final_rss - initial_rss
    rss_growth_percent = (rss_growth / initial_rss * 100) if initial_rss > 0 else 0
    
    # Find peak values
    peak_rss = max(rss_values) if rss_values else 0
    peak_vms = max(vms_values) if vms_values else 0
    peak_percent = max(percent_values) if percent_values else 0
    
    # Calculate average growth rate (MB per minute)
    if len(snapshots) > 1:
        time_span_minutes = (snapshots[-1].elapsed_seconds - snapshots[0].elapsed_seconds) / 60.0
        growth_rate_mb_per_min = rss_growth / time_span_minutes if time_span_minutes > 0 else 0
    else:
        growth_rate_mb_per_min = 0
    
    return {
        "initial_rss_mb": initial_rss,
        "final_rss_mb": final_rss,
        "peak_rss_mb": peak_rss,
        "peak_vms_mb": peak_vms,
        "peak_percent": peak_percent,
        "rss_growth_mb": rss_growth,
        "rss_growth_percent": rss_growth_percent,
        "growth_rate_mb_per_min": growth_rate_mb_per_min,
        "snapshot_count": len(snapshots),
        "time_span_minutes": (snapshots[-1].elapsed_seconds - snapshots[0].elapsed_seconds) / 60.0 if len(snapshots) > 1 else 0,
    }


async def run_memory_test(
    duration_minutes: float = 20.0,
    monitoring_interval_seconds: float = 60.0,
) -> Dict[str, Any]:
    """Run a long-duration recording test with memory monitoring."""
    duration_seconds = duration_minutes * 60.0
    
    logger.info("=" * 80)
    logger.info("Memory Monitoring Test for ParticipantRecorder")
    logger.info("=" * 80)
    logger.info(f"Duration: {duration_minutes} minutes ({duration_seconds} seconds)")
    logger.info(f"Monitoring interval: {monitoring_interval_seconds} seconds")
    
    # Enable profiling
    enable_profiling = os.getenv("LIVEKIT_ENABLE_PROFILING", "true").lower() == "true"
    if enable_profiling:
        tracemalloc.start()
        logger.info("Memory profiling enabled (tracemalloc + object counting)")
    else:
        logger.info("Memory profiling disabled")
    logger.info("=" * 80)
    
    # Get current process ID
    process_id = os.getpid()
    if HAS_PSUTIL:
        process = psutil.Process(process_id)
    else:
        process = process_id
    
    results: Dict[str, Any] = {
        "success": False,
        "errors": [],
        "warnings": [],
        "snapshots": [],
        "memory_analysis": {},
        "final_stats": {},
    }
    
    room = None
    recorder = None
    
    try:
        # Get room credentials
        if os.getenv("LIVEKIT_USE_PUMPFUN", "false").lower() == "true" and HAS_PUMPFUN:
            pumpfun = PumpFunService()
            stream_info = await pumpfun.get_random_live_stream(
                min_participants=1,
                exclude_nsfw=True,
            )
            if not stream_info:
                raise RecordingError("Failed to get stream from pump.fun")
            
            mint_id = stream_info.get("mint")
            if not mint_id:
                raise RecordingError("Selected stream has no mint ID")
            
            logger.info(f"Getting token for stream: {stream_info.get('name', 'Unknown')} ({mint_id})")
            token_str = await pumpfun.get_livestream_token(mint_id, role="viewer")
            if not token_str:
                raise RecordingError("Failed to get token from pump.fun")
            
            url = pumpfun.get_livekit_url()
            room_name = "pump.fun stream"
            logger.info(f"Using pump.fun stream: {stream_info.get('name', 'Unknown')}")
        else:
            url = os.getenv("LIVEKIT_URL")
            api_key = os.getenv("LIVEKIT_API_KEY")
            api_secret = os.getenv("LIVEKIT_API_SECRET")
            room_name = os.getenv("LIVEKIT_ROOM_NAME", "test-memory-room")
            
            if not url or not api_key or not api_secret:
                raise RecordingError("Missing LIVEKIT_URL, LIVEKIT_API_KEY, or LIVEKIT_API_SECRET")
            
            token_opts = livekit_api.CreateTokenOptions()
            token_opts.identity = "memory-test-client"
            token = livekit_api.AccessToken(api_key, api_secret, options=token_opts)
            token.grant_room_join(room_name)
            token_str = token.to_jwt()
        
        # Connect to room
        logger.info(f"Connecting to room: {room_name}")
        room = rtc.Room()
        await room.connect(url, token_str)
        logger.info("Connected to room")
        
        # Wait for participant
        logger.info("Waiting for participant...")
        participant = await wait_for_participant(room, timeout=30.0)
        if not participant:
            raise RecordingError("No participant found in room")
        
        logger.info(f"Participant found: {participant.identity}")
        
        # Wait for tracks
        has_video, has_audio = await wait_for_tracks(participant, timeout=20.0)
        logger.info(f"Tracks available - Video: {has_video}, Audio: {has_audio}")
        
        if not has_video and not has_audio:
            results["warnings"].append("No video or audio tracks found")
        
        # Create recorder
        recorder = ParticipantRecorder(room)
        logger.info("Recorder created")
        
        # Start recording
        logger.info(f"Starting recording for {duration_minutes} minutes...")
        await recorder.start_recording(participant.identity)
        
        start_time = time.time()
        snapshots: List[MemorySnapshot] = []
        
        # Start memory monitoring task
        monitoring_task = asyncio.create_task(
            monitor_memory(
                process,
                recorder,
                start_time,
                duration_seconds,
                monitoring_interval_seconds,
                snapshots,
                enable_profiling=enable_profiling,
            )
        )
        
        # Record for specified duration
        await asyncio.sleep(duration_seconds)
        
        # Wait for monitoring task to complete
        await monitoring_task
        
        # Get final stats
        stats = recorder.get_stats()
        results["final_stats"] = {
            "video_frames_recorded": stats.video_frames_recorded,
            "audio_frames_recorded": stats.audio_frames_recorded,
            "recording_duration_seconds": stats.recording_duration_seconds,
        }
        
        logger.info(f"Final stats: {results['final_stats']}")
        
        # Stop recording
        logger.info("Stopping recording...")
        output_file = tempfile.NamedTemporaryFile(suffix=".webm", delete=False).name
        await recorder.stop_recording(output_file)
        results["output_file"] = output_file
        results["output_file_size_bytes"] = os.path.getsize(output_file) if os.path.exists(output_file) else 0
        
        # Get final profiling snapshot
        if enable_profiling:
            final_tracemalloc_stats = get_tracemalloc_stats()
            final_object_counts = get_object_counts()
            results["final_tracemalloc_stats"] = final_tracemalloc_stats
            results["final_object_counts"] = final_object_counts
            
            if final_tracemalloc_stats:
                logger.info(
                    f"Final tracemalloc: {final_tracemalloc_stats.get('current_mb', 0):.1f}MB "
                    f"(peak: {final_tracemalloc_stats.get('peak_mb', 0):.1f}MB)"
                )
        
        # Analyze memory usage
        results["snapshots"] = [asdict(s) for s in snapshots]
        results["memory_analysis"] = analyze_memory_usage(snapshots)
        
        results["success"] = True
        
    except Exception as e:
        logger.error(f"Error during test: {e}", exc_info=True)
        results["errors"].append(str(e))
        results["success"] = False
    
    finally:
        if enable_profiling and tracemalloc.is_tracing():
            tracemalloc.stop()
        
        if recorder:
            try:
                await recorder.stop_recording()
            except:
                pass
        
        if room:
            try:
                await room.disconnect()
            except:
                pass
    
    return results


def print_memory_report(results: Dict[str, Any]) -> None:
    """Print a formatted memory usage report."""
    print("\n" + "=" * 80)
    print("Memory Usage Report")
    print("=" * 80)
    
    if not results.get("success"):
        print("❌ Test failed")
        if results.get("errors"):
            print("\nErrors:")
            for error in results["errors"]:
                print(f"  - {error}")
        return
    
    analysis = results.get("memory_analysis", {})
    snapshots = results.get("snapshots", [])
    
    print(f"\nTest Duration: {results['final_stats'].get('recording_duration_seconds', 0) / 60:.1f} minutes")
    print(f"Monitoring Intervals: {len(snapshots)} snapshots")
    
    if analysis:
        print("\nMemory Statistics:")
        print(f"  Initial RSS: {analysis.get('initial_rss_mb', 0):.1f} MB")
        print(f"  Final RSS: {analysis.get('final_rss_mb', 0):.1f} MB")
        print(f"  Peak RSS: {analysis.get('peak_rss_mb', 0):.1f} MB")
        print(f"  Peak VMS: {analysis.get('peak_vms_mb', 0):.1f} MB")
        print(f"  Peak %: {analysis.get('peak_percent', 0):.1f}%")
        
        print("\nMemory Growth:")
        growth_mb = analysis.get('rss_growth_mb', 0)
        growth_percent = analysis.get('rss_growth_percent', 0)
        growth_rate = analysis.get('growth_rate_mb_per_min', 0)
        
        print(f"  Total Growth: {growth_mb:+.1f} MB ({growth_percent:+.1f}%)")
        print(f"  Growth Rate: {growth_rate:+.2f} MB/minute")
        
        if abs(growth_mb) > 100 or abs(growth_rate) > 5:
            print("\n⚠️  WARNING: Significant memory growth detected!")
        elif abs(growth_mb) < 50 and abs(growth_rate) < 2:
            print("\n✓ Memory usage appears stable")
    
    print("\nTimeline (every 60 seconds):")
    print(f"{'Time':<10} {'RSS (MB)':<12} {'VMS (MB)':<12} {'%':<8} {'Video':<10} {'Audio':<10}")
    print("-" * 80)
    
    for snapshot in snapshots:
        elapsed_min = snapshot["elapsed_seconds"] / 60.0
        print(
            f"{elapsed_min:>6.1f}m  "
            f"{snapshot['rss_mb']:>10.1f}  "
            f"{snapshot['vms_mb']:>10.1f}  "
            f"{snapshot['percent']:>6.1f}  "
            f"{snapshot['video_frames']:>8}  "
            f"{snapshot['audio_frames']:>8}"
        )
    
    print("\nFinal Recording Stats:")
    stats = results.get("final_stats", {})
    print(f"  Video frames: {stats.get('video_frames_recorded', 0):,}")
    print(f"  Audio frames: {stats.get('audio_frames_recorded', 0):,}")
    print(f"  File size: {results.get('output_file_size_bytes', 0):,} bytes")
    print(f"  Output file: {results.get('output_file', 'N/A')}")
    
    # Print profiling results if available
    if results.get("final_tracemalloc_stats"):
        print("\n" + "=" * 80)
        print("Memory Profiling Results")
        print("=" * 80)
        
        tracemalloc_stats = results["final_tracemalloc_stats"]
        print(f"\nTracemalloc Statistics:")
        print(f"  Current traced: {tracemalloc_stats.get('current_mb', 0):.1f} MB")
        print(f"  Peak traced: {tracemalloc_stats.get('peak_mb', 0):.1f} MB")
        
        if tracemalloc_stats.get("top_allocations"):
            print(f"\n  Top 10 Allocations by Size:")
            for alloc in tracemalloc_stats["top_allocations"]:
                print(
                    f"    {alloc['rank']}. {alloc['filename']}:{alloc['lineno']} - "
                    f"{alloc['size_mb']:.2f} MB ({alloc['count']:,} allocations)"
                )
    
    if results.get("final_object_counts"):
        print(f"\n  Top 10 Object Types by Count:")
        for obj_type, count in list(results["final_object_counts"].items())[:10]:
            print(f"    {obj_type}: {count:,} objects")
    
    print("=" * 80)


async def main():
    """Main entry point."""
    duration_minutes = float(os.getenv("LIVEKIT_RECORDING_DURATION", "20.0"))
    monitoring_interval = float(os.getenv("LIVEKIT_MEMORY_INTERVAL", "60.0"))
    
    results = await run_memory_test(
        duration_minutes=duration_minutes,
        monitoring_interval_seconds=monitoring_interval,
    )
    
    print_memory_report(results)
    
    # Save results to JSON file
    output_file = Path(f"/tmp/memory_test_results_{int(time.time())}.json")
    with open(output_file, "w") as f:
        json.dump(results, f, indent=2)
    
    print(f"\nDetailed results saved to: {output_file}")
    
    if not results.get("success"):
        sys.exit(1)


if __name__ == "__main__":
    asyncio.run(main())




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rpc.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\trpc.proto\x12\rlivekit.proto\"7\n\x08RpcError\x12\x0c\n\x04\x63ode\x18\x01 \x02(\r\x12\x0f\n\x07message\x18\x02 \x02(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t\"\x91\x01\n\x11PerformRpcRequest\x12 \n\x18local_participant_handle\x18\x01 \x02(\x04\x12\x1c\n\x14\x64\x65stination_identity\x18\x02 \x02(\t\x12\x0e\n\x06method\x18\x03 \x02(\t\x12\x0f\n\x07payload\x18\x04 \x02(\t\x12\x1b\n\x13response_timeout_ms\x18\x05 \x01(\r\"L\n\x18RegisterRpcMethodRequest\x12 \n\x18local_participant_handle\x18\x01 \x02(\x04\x12\x0e\n\x06method\x18\x02 \x02(\t\"N\n\x1aUnregisterRpcMethodRequest\x12 \n\x18local_participant_handle\x18\x01 \x02(\x04\x12\x0e\n\x06method\x18\x02 \x02(\t\"\x96\x01\n\"RpcMethodInvocationResponseRequest\x12 \n\x18local_participant_handle\x18\x01 \x02(\x04\x12\x15\n\rinvocation_id\x18\x02 \x02(\x04\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12&\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x17.livekit.proto.RpcError\"&\n\x12PerformRpcResponse\x12\x10\n\x08\x61sync_id\x18\x01 \x02(\x04\"\x1b\n\x19RegisterRpcMethodResponse\"\x1d\n\x1bUnregisterRpcMethodResponse\"4\n#RpcMethodInvocationResponseResponse\x12\r\n\x05\x65rror\x18\x01 \x01(\t\"_\n\x12PerformRpcCallback\x12\x10\n\x08\x61sync_id\x18\x01 \x02(\x04\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12&\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x17.livekit.proto.RpcError\"\xbe\x01\n\x18RpcMethodInvocationEvent\x12 \n\x18local_participant_handle\x18\x01 \x02(\x04\x12\x15\n\rinvocation_id\x18\x02 \x02(\x04\x12\x0e\n\x06method\x18\x03 \x02(\t\x12\x12\n\nrequest_id\x18\x04 \x02(\t\x12\x17\n\x0f\x63\x61ller_identity\x18\x05 \x02(\t\x12\x0f\n\x07payload\x18\x06 \x02(\t\x12\x1b\n\x13response_timeout_ms\x18\x07 \x02(\rB\x10\xaa\x02\rLiveKit.Proto')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'rpc_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\252\002\rLiveKit.Proto'
  _globals['_RPCERROR']._serialized_start=28
  _globals['_RPCERROR']._serialized_end=83
  _globals['_PERFORMRPCREQUEST']._serialized_start=86
  _globals['_PERFORMRPCREQUEST']._serialized_end=231
  _globals['_REGISTERRPCMETHODREQUEST']._serialized_start=233
  _globals['_REGISTERRPCMETHODREQUEST']._serialized_end=309
  _globals['_UNREGISTERRPCMETHODREQUEST']._serialized_start=311
  _globals['_UNREGISTERRPCMETHODREQUEST']._serialized_end=389
  _globals['_RPCMETHODINVOCATIONRESPONSEREQUEST']._serialized_start=392
  _globals['_RPCMETHODINVOCATIONRESPONSEREQUEST']._serialized_end=542
  _globals['_PERFORMRPCRESPONSE']._serialized_start=544
  _globals['_PERFORMRPCRESPONSE']._serialized_end=582
  _globals['_REGISTERRPCMETHODRESPONSE']._serialized_start=584
  _globals['_REGISTERRPCMETHODRESPONSE']._serialized_end=611
  _globals['_UNREGISTERRPCMETHODRESPONSE']._serialized_start=613
  _globals['_UNREGISTERRPCMETHODRESPONSE']._serialized_end=642
  _globals['_RPCMETHODINVOCATIONRESPONSERESPONSE']._serialized_start=644
  _globals['_RPCMETHODINVOCATIONRESPONSERESPONSE']._serialized_end=696
  _globals['_PERFORMRPCCALLBACK']._serialized_start=698
  _globals['_PERFORMRPCCALLBACK']._serialized_end=793
  _globals['_RPCMETHODINVOCATIONEVENT']._serialized_start=796
  _globals['_RPCMETHODINVOCATIONEVENT']._serialized_end=986
# @@protoc_insertion_point(module_scope)



# Copyright 2023 LiveKit, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, cast
import asyncio

from ._ffi_client import FfiHandle, FfiClient
from ._proto import e2ee_pb2 as proto_e2ee
from ._proto import ffi_pb2 as proto_ffi
from ._proto import track_pb2 as proto_track
from .track import Track, LocalTrack, RemoteTrack


class TrackPublication:
    def __init__(self, owned_info: proto_track.OwnedTrackPublication):
        self._info = owned_info.info
        self._track: Optional[Track] = None
        self._ffi_handle = FfiHandle(owned_info.handle.id)

    @property
    def track(self) -> Optional[Track]:
        return self._track

    @property
    def sid(self) -> str:
        return self._info.sid

    @property
    def name(self) -> str:
        return self._info.name

    @property
    def kind(self) -> proto_track.TrackKind.ValueType:
        return self._info.kind

    @property
    def source(self) -> proto_track.TrackSource.ValueType:
        return self._info.source

    @property
    def simulcasted(self) -> bool:
        return self._info.simulcasted

    @property
    def width(self) -> int:
        return self._info.width

    @property
    def height(self) -> int:
        return self._info.height

    @property
    def mime_type(self) -> str:
        return self._info.mime_type

    @property
    def muted(self) -> bool:
        return self._info.muted

    @property
    def encryption_type(self) -> proto_e2ee.EncryptionType.ValueType:
        return self._info.encryption_type

    @property
    def audio_features(self) -> List[proto_track.AudioTrackFeature.ValueType]:
        return list(self._info.audio_features)


class LocalTrackPublication(TrackPublication):
    def __init__(self, owned_info: proto_track.OwnedTrackPublication):
        super().__init__(owned_info)
        self._first_subscription: asyncio.Future[None] = asyncio.Future()

    @property
    def track(self) -> Optional[LocalTrack]:
        return cast(Optional[LocalTrack], self._track)

    async def wait_for_subscription(self) -> None:
        await asyncio.shield(self._first_subscription)

    def __repr__(self) -> str:
        return f"rtc.LocalTrackPublication(sid={self.sid}, name={self.name}, kind={self.kind}, source={self.source})"


class RemoteTrackPublication(TrackPublication):
    def __init__(self, owned_info: proto_track.OwnedTrackPublication):
        super().__init__(owned_info)
        self._subscribed = False

    @property
    def track(self) -> Optional[RemoteTrack]:
        return cast(Optional[RemoteTrack], self._track)

    @property
    def subscribed(self) -> bool:
        return self._subscribed

    def set_subscribed(self, subscribed: bool):
        req = proto_ffi.FfiRequest()
        req.set_subscribed.subscribe = subscribed
        req.set_subscribed.publication_handle = self._ffi_handle.handle
        FfiClient.instance.request(req)

    def __repr__(self) -> str:
        return f"rtc.RemoteTrackPublication(sid={self.sid}, name={self.name}, kind={self.kind}, source={self.source})"
